import asyncio
from pathlib import Path, PurePath
from typing import Dict, List, Set, Tuple

import click
from click.core import Context
from intervaltree import IntervalTree
from rich.panel import Panel

from woke.ast.nodes import AstSolc
from woke.compile import SolcOutput, SolidityCompiler
from woke.compile.solc_frontend.input_data_model import SolcOutputSelectionEnum
from woke.config import WokeConfig

from ..ast.ir.meta.source_unit import SourceUnit
from ..ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from ..ast.ir.utils import IrInitTuple
from ..compile.compilation_unit import CompilationUnit
from ..compile.solc_frontend import SolcOutputErrorSeverityEnum
from ..utils.file_utils import is_relative_to
from .console import console


@click.command(name="compile")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--parse", is_flag=True, default=False, help="Also try to parse the generated AST."
)
@click.option(
    "--no-artifacts", is_flag=True, default=False, help="Do not write build artifacts."
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force recompile the project without previous build artifacts.",
)
@click.pass_context
def run_compile(
    ctx: Context, paths: Tuple[str], parse: bool, no_artifacts: bool, force: bool
) -> None:
    """Compile the project."""
    config = WokeConfig(woke_root_path=ctx.obj["woke_root_path"])
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    sol_files: Set[Path] = set()
    if len(paths) == 0:
        for file in config.project_root_path.rglob("**/*.sol"):
            if (
                not any(
                    is_relative_to(file, p) for p in config.compiler.solc.ignore_paths
                )
                and file.is_file()
            ):
                sol_files.add(file)
    else:
        for p in paths:
            path = Path(p)
            if path.is_file():
                if not path.match("*.sol"):
                    raise ValueError(f"Argument `{p}` is not a Solidity file.")
                sol_files.add(path)
            elif path.is_dir():
                for file in path.rglob("**/*.sol"):
                    if (
                        not any(
                            is_relative_to(file, p)
                            for p in config.compiler.solc.ignore_paths
                        )
                        and file.is_file()
                    ):
                        sol_files.add(file)
            else:
                raise ValueError(f"Argument `{p}` is not a file or directory.")

    compiler = SolidityCompiler(config)
    # TODO Allow choosing build artifacts subset in compile subcommand
    outputs: List[Tuple[CompilationUnit, SolcOutput]] = asyncio.run(
        compiler.compile(
            sol_files,
            [SolcOutputSelectionEnum.ALL],
            write_artifacts=(not no_artifacts),
            reuse_latest_artifacts=(not force),
            maximize_compilation_units=True,
        )
    )

    errored = False
    for _, output in outputs:
        for error in output.errors:
            if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                errored = True
            if error.formatted_message is not None:
                console.print(Panel(error.formatted_message, highlight=True))
            else:
                console.print(Panel(error.message, highlight=True))

    if parse and not errored:
        processed_files: Set[Path] = set()
        reference_resolver = ReferenceResolver()
        interval_trees: Dict[Path, IntervalTree] = {}
        source_units: Dict[Path, SourceUnit] = {}

        for cu, output in outputs:
            for source_unit_name, info in output.sources.items():
                path = cu.source_unit_name_to_path(PurePath(source_unit_name))
                ast = AstSolc.parse_obj(info.ast)

                reference_resolver.index_nodes(ast, path, cu.hash)

                if path in processed_files:
                    continue
                processed_files.add(path)
                interval_trees[path] = IntervalTree()

                init = IrInitTuple(
                    path,
                    path.read_bytes(),
                    cu,
                    interval_trees[path],
                    reference_resolver,
                )
                source_units[path] = SourceUnit(init, ast)

        reference_resolver.run_post_process_callbacks(
            CallbackParams(interval_trees=interval_trees, source_units=source_units)
        )



from __future__ import annotations

import asyncio
import difflib
import enum
import logging
import re
import threading
from collections import deque
from copy import deepcopy
from functools import lru_cache
from pathlib import Path, PurePath
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Deque,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from woke.compile.exceptions import CompilationError

from ..analysis.detectors import detect
from ..core.solidity_version import SolidityVersionRange, SolidityVersionRanges
from ..utils.file_utils import is_relative_to

if TYPE_CHECKING:
    from .server import LspServer

import networkx as nx
from intervaltree import IntervalTree

from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.reference_resolver import CallbackParams, ReferenceResolver
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstSolc
from woke.compile import SolcOutput, SolcOutputSelectionEnum
from woke.compile.compilation_unit import CompilationUnit
from woke.compile.compiler import SolidityCompiler
from woke.compile.solc_frontend import SolcOutputError, SolcOutputErrorSeverityEnum
from woke.config import WokeConfig
from woke.lsp.document_sync import (
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
)
from woke.lsp.utils.uri import path_to_uri, uri_to_path

from ..svm import SolcVersionManager
from .common_structures import (
    CreateFilesParams,
    DeleteFile,
    DeleteFilesParams,
    Diagnostic,
    DiagnosticRelatedInformation,
    DiagnosticSeverity,
    DocumentUri,
    Location,
    MessageType,
    Position,
    Range,
    RenameFilesParams,
)

logger = logging.getLogger(__name__)


ENCODING = "utf-16-le"


def _binary_search(lines: List[Tuple[bytes, int]], x: int) -> int:
    l = 0
    r = len(lines)

    while l < r:
        mid = l + (r - l) // 2
        if lines[mid][1] < x + 1:
            l = mid + 1
        else:
            r = mid

    return l - 1


def _out_edge_bfs(cu: CompilationUnit, start: Iterable[Path], out: Set[Path]) -> None:
    processed: Set[PurePath] = set()
    for path in start:
        processed.update(cu.path_to_source_unit_names(path))
    out.update(start)

    queue: Deque[PurePath] = deque(processed)
    while len(queue):
        node = queue.pop()
        for out_edge in cu.graph.out_edges(node):
            from_, to = out_edge
            if to not in processed:
                processed.add(to)
                queue.append(to)
                out.add(cu.source_unit_name_to_path(to))


class VersionedFile(NamedTuple):
    text: str
    version: Optional[int]


class CustomFileChangeCommand(str, enum.Enum):
    FORCE_RECOMPILE = "force_recompile"
    FORCE_RERUN_DETECTORS = "force_rerun_detectors"


class LspCompiler:
    __config: WokeConfig
    __svm: SolcVersionManager
    __server: LspServer
    __file_changes_queue: asyncio.Queue
    __diagnostic_queue: asyncio.Queue
    __discovered_files: Set[Path]
    __deleted_files: Set[Path]
    __opened_files: Dict[Path, VersionedFile]
    __modified_files: Set[Path]
    __force_compile_files: Set[Path]
    __compiler: SolidityCompiler
    __output_contents: Dict[Path, VersionedFile]
    __compilation_errors: Dict[Path, Set[Diagnostic]]
    __last_successful_compilation_contents: Dict[Path, VersionedFile]
    __interval_trees: Dict[Path, IntervalTree]
    __source_units: Dict[Path, SourceUnit]
    __last_compilation_interval_trees: Dict[Path, IntervalTree]
    __last_compilation_source_units: Dict[Path, SourceUnit]
    __line_indexes: Dict[Path, List[Tuple[bytes, int]]]
    __perform_files_discovery: bool
    __force_run_detectors: bool

    __ir_reference_resolver: ReferenceResolver

    __output_ready: asyncio.Event

    def __init__(
        self,
        server: LspServer,
        diagnostic_queue: asyncio.Queue,
        perform_files_discovery: bool,
    ):
        self.__server = server
        self.__file_changes_queue = asyncio.Queue()
        self.__diagnostic_queue = diagnostic_queue
        self.__stop_event = threading.Event()
        self.__discovered_files = set()
        self.__deleted_files = set()
        self.__opened_files = {}
        self.__modified_files = set()
        self.__force_compile_files = set()
        self.__interval_trees = {}
        self.__source_units = {}
        self.__last_compilation_interval_trees = {}
        self.__last_compilation_source_units = {}
        self.__line_indexes = {}
        self.__output_contents = dict()
        self.__compilation_errors = dict()
        self.__last_successful_compilation_contents = dict()
        self.__output_ready = asyncio.Event()
        self.__perform_files_discovery = perform_files_discovery
        self.__force_run_detectors = False

        self.__ir_reference_resolver = ReferenceResolver()

    async def run(self, config: WokeConfig):
        self.__config = config
        self.__svm = SolcVersionManager(config)
        self.__compiler = SolidityCompiler(config)
        await self.__compilation_loop()

    @property
    def output_ready(self) -> asyncio.Event:
        return self.__output_ready

    @property
    def ir_reference_resolver(self) -> ReferenceResolver:
        return self.__ir_reference_resolver

    @property
    def interval_trees(self) -> Dict[Path, IntervalTree]:
        return self.__interval_trees

    @property
    def source_units(self) -> Dict[Path, SourceUnit]:
        return self.__source_units

    @property
    def last_compilation_interval_trees(self) -> Dict[Path, IntervalTree]:
        return self.__last_compilation_interval_trees

    @property
    def last_compilation_source_units(self) -> Dict[Path, SourceUnit]:
        return self.__last_compilation_source_units

    @lru_cache(maxsize=128)
    def _compute_diff_interval_tree(
        self, a: VersionedFile, b: VersionedFile
    ) -> IntervalTree:
        seq_matcher = difflib.SequenceMatcher(None, a.text, b.text)
        interval_tree = IntervalTree()
        for tag, i1, i2, j1, j2 in seq_matcher.get_opcodes():
            if tag == "equal":
                continue
            interval_tree.addi(i1, i2 + 1, (tag, j1, j2 + 1))
        return interval_tree

    def get_last_compilation_forward_changes(
        self, path: Path
    ) -> Optional[IntervalTree]:
        if path not in self.__last_successful_compilation_contents:
            return None
        return self._compute_diff_interval_tree(
            self.__last_successful_compilation_contents[path],
            self.get_compiled_file(path),
        )

    def get_last_compilation_backward_changes(
        self, path: Path
    ) -> Optional[IntervalTree]:
        if path not in self.__last_successful_compilation_contents:
            return None
        return self._compute_diff_interval_tree(
            self.get_compiled_file(path),
            self.__last_successful_compilation_contents[path],
        )

    async def add_change(
        self,
        change: Union[
            DidOpenTextDocumentParams,
            DidChangeTextDocumentParams,
            DidCloseTextDocumentParams,
            CreateFilesParams,
            RenameFilesParams,
            DeleteFilesParams,
        ],
    ) -> None:
        self.output_ready.clear()
        await self.__file_changes_queue.put(change)

    async def force_recompile(self) -> None:
        self.__output_ready.clear()
        await self.__file_changes_queue.put(CustomFileChangeCommand.FORCE_RECOMPILE)

    async def force_rerun_detectors(self) -> None:
        await self.__file_changes_queue.put(
            CustomFileChangeCommand.FORCE_RERUN_DETECTORS
        )

    def get_compiled_file(self, file: Union[Path, str]) -> VersionedFile:
        if isinstance(file, str):
            file = uri_to_path(file)
        if file not in self.__output_contents:
            self.__output_contents[file] = VersionedFile(
                file.read_bytes().decode(encoding="utf-8"), None
            )
        return self.__output_contents[file]

    def get_line_pos_from_byte_offset(
        self, file: Path, byte_offset: int
    ) -> Tuple[int, int]:
        if file not in self.__line_indexes:
            self.__setup_line_index(file)

        encoded_lines = self.__line_indexes[file]
        line_num = _binary_search(encoded_lines, byte_offset)
        line_data, prefix_sum = encoded_lines[line_num]
        line_offset = byte_offset - prefix_sum
        return line_num, line_offset

    def get_range_from_byte_offsets(
        self, file: Path, byte_offsets: Tuple[int, int]
    ) -> Range:
        start_line, start_column = self.get_line_pos_from_byte_offset(
            file, byte_offsets[0]
        )
        end_line, end_column = self.get_line_pos_from_byte_offset(file, byte_offsets[1])

        return Range(
            start=Position(line=start_line, character=start_column),
            end=Position(line=end_line, character=end_column),
        )

    def get_byte_offset_from_line_pos(self, file: Path, line: int, col: int) -> int:
        if file not in self.__line_indexes:
            self.__setup_line_index(file)

        encoded_lines = self.__line_indexes[file]
        line_bytes, prefix = encoded_lines[line]
        line_offset = len(line_bytes.decode("utf-8")[:col].encode("utf-8"))
        return prefix + line_offset

    async def _handle_change(
        self,
        change: Union[
            DidOpenTextDocumentParams,
            DidCloseTextDocumentParams,
            DidChangeTextDocumentParams,
            None,
        ],
    ) -> None:
        if isinstance(change, CustomFileChangeCommand):
            if change == CustomFileChangeCommand.FORCE_RECOMPILE:
                for file in self.__discovered_files:
                    # clear diagnostics
                    await self.__diagnostic_queue.put((file, set()))

                self.__discovered_files.clear()
                if self.__perform_files_discovery:
                    for file in self.__config.project_root_path.rglob("**/*.sol"):
                        if not self.__file_ignored(file) and file.is_file():
                            self.__discovered_files.add(file.resolve())

                for file in self.__opened_files.keys():
                    if not self.__file_ignored(file):
                        self.__discovered_files.add(file)

                self.__force_compile_files.update(self.__discovered_files)
            elif change == CustomFileChangeCommand.FORCE_RERUN_DETECTORS:
                self.__force_run_detectors = True
        elif isinstance(change, CreateFilesParams):
            for file in change.files:
                path = uri_to_path(file.uri)
                if (
                    path not in self.__discovered_files
                    and not self.__file_ignored(path)
                    and path.is_file()
                    and path.suffix == ".sol"
                ):
                    self.__discovered_files.add(path)
                    self.__force_compile_files.add(path)
        elif isinstance(change, RenameFilesParams):
            for rename in change.files:
                old_path = uri_to_path(rename.old_uri)
                self.__deleted_files.add(old_path)
                self.__discovered_files.discard(old_path)

                new_path = uri_to_path(rename.new_uri)
                if (
                    new_path not in self.__discovered_files
                    and not self.__file_ignored(new_path)
                    and new_path.is_file()
                    and new_path.suffix == ".sol"
                ):
                    self.__discovered_files.add(new_path)
                    self.__force_compile_files.add(new_path)
        elif isinstance(change, DeleteFilesParams):
            for delete in change.files:
                path = uri_to_path(delete.uri)
                self.__deleted_files.add(path)
                self.__discovered_files.discard(path)
        elif isinstance(change, DidOpenTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            self.__opened_files[path] = VersionedFile(
                change.text_document.text, change.text_document.version
            )
            if (
                path not in self.__discovered_files
                and not self.__file_ignored(path)
                and path.is_file()
                and path.suffix == ".sol"
            ):
                self.__discovered_files.add(path)
                self.__force_compile_files.add(path)
            elif change.text_document.text != self.get_compiled_file(path).text:
                self.__force_compile_files.add(path)
            else:
                self.__output_contents[path] = self.__opened_files[path]

        elif isinstance(change, DidCloseTextDocumentParams):
            pass
        elif isinstance(change, DidChangeTextDocumentParams):
            path = uri_to_path(change.text_document.uri).resolve()
            self.__modified_files.add(path)

            for content_change in change.content_changes:
                start = content_change.range.start
                end = content_change.range.end

                # str.splitlines() removes empty lines => cannot be used
                # str.split() removes separators => cannot be used
                tmp_lines = re.split(r"(\r?\n)", self.__opened_files[path].text)
                tmp_lines2: List[str] = []
                for line in tmp_lines:
                    if line in {"\r\n", "\n"}:
                        tmp_lines2[-1] += line
                    else:
                        tmp_lines2.append(line)

                lines: List[bytearray] = [
                    bytearray(line.encode(ENCODING)) for line in tmp_lines2
                ]

                if start.line == end.line:
                    line = lines[start.line]
                    line[start.character * 2 : end.character * 2] = b""
                    line[
                        start.character * 2 : start.character * 2
                    ] = content_change.text.encode(ENCODING)
                else:
                    start_line = lines[start.line]
                    end_line = lines[end.line]
                    start_line[start.character * 2 :] = content_change.text.encode(
                        ENCODING
                    )
                    end_line[: end.character * 2] = b""

                    for i in range(start.line + 1, end.line):
                        lines[i] = bytearray(b"")

                self.__opened_files[path] = VersionedFile(
                    "".join(line.decode(ENCODING) for line in lines),
                    change.text_document.version,
                )
        else:
            raise Exception("Unknown change type")

    async def __compile(
        self,
        files_to_compile: AbstractSet[Path],
        full_compile: bool = True,
    ) -> None:
        target_version = self.__config.compiler.solc.target_version
        min_version = self.__config.min_solidity_version
        max_version = self.__config.max_solidity_version
        if target_version is not None and target_version < min_version:
            await self.__server.log_message(
                f"The minimum supported version of Solidity is {min_version}. Version {target_version} is selected in settings.",
                MessageType.ERROR,
            )
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return
        if target_version is not None and target_version > max_version:
            await self.__server.log_message(
                f"The maximum supported version of Solidity is {max_version}. Version {target_version} is selected in settings.",
                MessageType.ERROR,
            )
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        try:
            if full_compile:
                graph = self.__compiler.build_graph(
                    self.__discovered_files,
                    {path: info.text for path, info in self.__opened_files.items()},
                    True,
                )
            else:
                graph = self.__compiler.build_graph(
                    files_to_compile,
                    {path: info.text for path, info in self.__opened_files.items()},
                    True,
                )
        except CompilationError as e:
            await self.__server.log_message(str(e), MessageType.ERROR)
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        compilation_units = self.__compiler.build_compilation_units_maximize(graph)

        # filter out only compilation units that need to be compiled
        compilation_units = [
            cu
            for cu in compilation_units
            if (cu.files & files_to_compile)
            or cu.contains_unresolved_file(self.__deleted_files, self.__config)
        ]
        build_settings = self.__compiler.create_build_settings(
            [SolcOutputSelectionEnum.AST]
        )

        # optimization - merge compilation units that can be compiled together
        if all(len(cu.versions) for cu in compilation_units):
            compilation_units = sorted(
                compilation_units, key=lambda cu: cu.versions.version_ranges[0].lower
            )

            merged_compilation_units: List[CompilationUnit] = []
            source_unit_names: Set = set()
            versions = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )

            for cu in compilation_units:
                if versions & cu.versions:
                    source_unit_names |= cu.source_unit_names
                    versions &= cu.versions
                else:
                    merged_compilation_units.append(
                        CompilationUnit(
                            graph.subgraph(source_unit_names).copy(),
                            versions,
                        )
                    )
                    source_unit_names = set(cu.source_unit_names)
                    versions = cu.versions

            merged_compilation_units.append(
                CompilationUnit(
                    graph.subgraph(source_unit_names).copy(),
                    versions,
                )
            )

            compilation_units = merged_compilation_units

        target_versions = []
        skipped_compilation_units = []
        for compilation_unit in compilation_units:
            target_version = self.__config.compiler.solc.target_version
            if target_version is not None:
                if target_version not in compilation_unit.versions:
                    await self.__server.log_message(
                        f"Unable to compile the following files with solc version `{target_version}` set in config:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
            else:
                # use the latest matching version
                matching_versions = [
                    version
                    for version in reversed(self.__svm.list_all())
                    if version in compilation_unit.versions
                ]
                if len(matching_versions) == 0:
                    await self.__server.log_message(
                        f"Unable to find a matching version of Solidity for the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
                try:
                    target_version = next(
                        version
                        for version in matching_versions
                        if version <= max_version
                    )
                except StopIteration:
                    await self.__server.log_message(
                        f"The maximum supported version of Solidity is {max_version}, unable to compile the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue

                if target_version < min_version:
                    await self.__server.log_message(
                        f"The minimum supported version of Solidity is {min_version}, unable to compile the following files:\n"
                        + "\n".join(
                            path_to_uri(path) for path in compilation_unit.files
                        ),
                        MessageType.ERROR,
                    )
                    skipped_compilation_units.append(compilation_unit)
                    continue
            target_versions.append(target_version)

        for version in set(target_versions):
            if not self.__svm.installed(version):
                progress_token = await self.__server.progress_begin(
                    "Downloading", f"solc {version}", 0
                )
                if progress_token is not None:

                    async def on_progress(downloaded: int, total: int) -> None:
                        assert progress_token is not None
                        await self.__server.progress_report(
                            progress_token,
                            f"solc {version}",
                            (100 * downloaded) // total,
                        )

                    await self.__svm.install(version, progress=on_progress)
                    await self.__server.progress_end(progress_token)
                else:
                    await self.__svm.install(version)

        for compilation_unit in skipped_compilation_units:
            for file in compilation_unit.files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
                self.__interval_trees.pop(file, None)
                self.__source_units.pop(file, None)
            compilation_units.remove(compilation_unit)

        progress_token = await self.__server.progress_begin(
            "Compiling", f"0/{len(compilation_units)}", 0
        )

        tasks = []
        for compilation_unit, target_version in zip(compilation_units, target_versions):
            task = self.__server.create_task(
                self.__compiler.compile_unit_raw(
                    compilation_unit,
                    target_version,
                    build_settings,
                )
            )
            tasks.append(task)

        # wait for compilation of all compilation units
        try:
            ret = await asyncio.gather(*tasks)
        except Exception as e:
            for task in tasks:
                task.cancel()
            await self.__server.log_message(str(e), MessageType.ERROR)
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        errors_without_location: Set[SolcOutputError] = set()
        errors_per_file: Dict[Path, Set[Diagnostic]] = deepcopy(
            self.__compilation_errors
        )

        for cu, solc_output in zip(compilation_units, ret):
            for file in cu.files:
                errors_per_file[file] = set()
            for error in solc_output.errors:
                if error.source_location is None:
                    errors_without_location.add(error)

        for error in errors_without_location:
            if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                error_type = MessageType.ERROR
            elif error.severity == SolcOutputErrorSeverityEnum.WARNING:
                error_type = MessageType.WARNING
            elif error.severity == SolcOutputErrorSeverityEnum.INFO:
                error_type = MessageType.INFO
            else:
                error_type = MessageType.LOG
            await self.__server.show_message(error.message, error_type)
            await self.__server.log_message(error.message, error_type)

        if (
            any(
                e.severity == SolcOutputErrorSeverityEnum.ERROR
                for e in errors_without_location
            )
            > 0
        ):
            for file in self.__discovered_files:
                # clear diagnostics
                await self.__diagnostic_queue.put((file, set()))
            self.__interval_trees.clear()
            self.__source_units.clear()
            return

        files_to_recompile = set(files_to_compile)
        processed_files: Set[Path] = set()

        for deleted_file in self.__deleted_files:
            await self.__diagnostic_queue.put((deleted_file, []))
            if deleted_file in self.__source_units:
                self.__ir_reference_resolver.run_destroy_callbacks(deleted_file)
                self.__source_units.pop(deleted_file)

        for cu_index, (cu, solc_output) in enumerate(zip(compilation_units, ret)):
            for file in cu.files:
                if file in self.__line_indexes:
                    self.__line_indexes.pop(file)
                if file in self.__opened_files:
                    self.__output_contents[file] = self.__opened_files[file]

            errored_files: Set[Path] = set()

            for error in solc_output.errors:
                if error.source_location is not None:
                    path = cu.source_unit_name_to_path(
                        PurePath(error.source_location.file)
                    )
                    errors_per_file[path].add(
                        self.__solc_error_to_diagnostic(error, path)
                    )
                    if error.severity == SolcOutputErrorSeverityEnum.ERROR:
                        errored_files.add(path)

            _out_edge_bfs(cu, errored_files, errored_files)

            # files requested to be compiled and files that import these files (even indirectly)
            recompiled_files: Set[Path] = set()
            _out_edge_bfs(cu, files_to_compile & cu.files, recompiled_files)

            for file in errored_files:
                files_to_recompile.discard(file)
                # an error occurred during compilation
                # AST still may be provided, but it must NOT be parsed (pydantic model is not defined for this case)
                if file in self.__source_units:
                    self.__ir_reference_resolver.run_destroy_callbacks(file)
                    self.__source_units.pop(file)
                if file in self.__interval_trees:
                    self.__interval_trees.pop(file)

            if len(errored_files) == 0:
                for source_unit_name, raw_ast in solc_output.sources.items():
                    path = cu.source_unit_name_to_path(PurePath(source_unit_name))
                    if path in errored_files or raw_ast.ast is None:
                        continue
                    ast = AstSolc.parse_obj(raw_ast.ast)

                    self.__ir_reference_resolver.index_nodes(ast, path, cu.hash)

                    files_to_recompile.discard(path)
                    if (
                        path in self.__source_units and path not in recompiled_files
                    ) or path in processed_files:
                        continue
                    processed_files.add(path)

                    interval_tree = IntervalTree()
                    init = IrInitTuple(
                        path,
                        self.get_compiled_file(path).text.encode("utf-8"),
                        cu,
                        interval_tree,
                        self.__ir_reference_resolver,
                    )
                    self.__ir_reference_resolver.run_destroy_callbacks(path)
                    self.__source_units[path] = SourceUnit(init, ast)
                    self.__interval_trees[path] = interval_tree

                    self.__last_compilation_source_units[path] = self.__source_units[
                        path
                    ]
                    self.__last_compilation_interval_trees[
                        path
                    ] = self.__interval_trees[path]

                    if path in self.__opened_files:
                        self.__last_successful_compilation_contents[
                            path
                        ] = self.__opened_files[path]
                    else:
                        self.__last_successful_compilation_contents[
                            path
                        ] = VersionedFile(
                            cu.graph.nodes[PurePath(source_unit_name)]["content"], None
                        )

            self.__ir_reference_resolver.run_post_process_callbacks(
                CallbackParams(
                    interval_trees=self.__interval_trees,
                    source_units=self.__source_units,
                )
            )

            if progress_token is not None:
                await self.__server.progress_report(
                    progress_token,
                    f"{cu_index + 1}/{len(compilation_units)}",
                    ((cu_index + 1) * 100) // len(compilation_units),
                )

        if progress_token is not None:
            await self.__server.progress_end(progress_token)

        self.__compilation_errors = deepcopy(errors_per_file)

        # diagnostics (compiler errors and warnings + detections) are sent in this function
        await self.__run_detectors()

        if len(files_to_recompile) > 0:
            # avoid infinite recursion
            if files_to_recompile != files_to_compile or full_compile:
                await self.__compile(files_to_recompile, False)

    async def __run_detectors(self):
        errors_per_file: Dict[Path, Set[Diagnostic]] = deepcopy(
            self.__compilation_errors
        )

        if self.__config.lsp.detectors.enable:
            for detection in detect(self.__config, self.__source_units):
                file = detection.result.ir_node.file
                if detection.result.related_info is not None:
                    related_info = [
                        DiagnosticRelatedInformation(
                            location=Location(
                                uri=DocumentUri(path_to_uri(info.ir_node.file)),
                                range=self.get_range_from_byte_offsets(
                                    info.ir_node.file, info.ir_node.byte_location
                                ),
                            ),
                            message=info.message,
                        )
                        for info in detection.result.related_info
                    ]
                else:
                    related_info = None

                if file not in errors_per_file:
                    errors_per_file[file] = set()
                errors_per_file[file].add(
                    Diagnostic(
                        range=self.get_range_from_byte_offsets(
                            file, detection.result.ir_node.byte_location
                        ),
                        severity=DiagnosticSeverity.WARNING,
                        source="Woke",
                        message=detection.result.message,
                        code=detection.code,
                        related_information=related_info,
                    )
                )

        for path, errors in errors_per_file.items():
            await self.__diagnostic_queue.put((path, errors))

    async def __compilation_loop(self):
        if self.__perform_files_discovery:
            # perform Solidity files discovery
            for file in self.__config.project_root_path.rglob("**/*.sol"):
                if not self.__file_ignored(file) and file.is_file():
                    self.__discovered_files.add(file.resolve())

            # perform initial compilation
            await self.__compile(self.__discovered_files)

        if self.__file_changes_queue.empty():
            self.output_ready.set()

        while True:
            change = await self.__file_changes_queue.get()
            while True:
                await self._handle_change(change)
                try:
                    change = self.__file_changes_queue.get_nowait()
                except asyncio.QueueEmpty:
                    break

            # run the compilation
            if (
                len(self.__force_compile_files) > 0
                or len(self.__modified_files) > 0
                or len(self.__deleted_files) > 0
            ):
                await self.__compile(
                    self.__force_compile_files.union(self.__modified_files)
                )

                self.__force_compile_files.clear()
                self.__modified_files.clear()
                self.__deleted_files.clear()

            if self.__force_run_detectors:
                await self.__run_detectors()
                self.__force_run_detectors = False

            if self.__file_changes_queue.empty():
                self.output_ready.set()

    def __setup_line_index(self, file: Path):
        content = self.get_compiled_file(file).text
        tmp_lines = re.split(r"(\r?\n)", content)
        lines: List[str] = []
        for line in tmp_lines:
            if line in {"\r\n", "\n"}:
                lines[-1] += line
            else:
                lines.append(line)

        # UTF-8 encoded lines with prefix length
        encoded_lines: List[Tuple[bytes, int]] = []
        prefix_sum = 0
        for line in lines:
            encoded_line = line.encode("utf-8")
            encoded_lines.append((encoded_line, prefix_sum))
            prefix_sum += len(encoded_line)
        self.__line_indexes[file] = encoded_lines

    def __solc_error_to_diagnostic(
        self, error: SolcOutputError, path: Path
    ) -> Diagnostic:
        assert error.source_location is not None
        if error.severity == SolcOutputErrorSeverityEnum.ERROR:
            severity = DiagnosticSeverity.ERROR
        elif error.severity == SolcOutputErrorSeverityEnum.WARNING:
            severity = DiagnosticSeverity.WARNING
        else:
            severity = DiagnosticSeverity.INFORMATION

        if error.source_location.start >= 0 and error.source_location.end >= 0:
            range_ = self.get_range_from_byte_offsets(
                path, (error.source_location.start, error.source_location.end)
            )
        else:
            range_ = Range(
                start=Position(line=0, character=0),
                end=Position(line=0, character=0),
            )

        return Diagnostic(
            range=range_,
            severity=severity,
            code=error.error_code,
            source="Woke(solc)",
            message=error.message,
        )

    def __file_ignored(self, path: Path) -> bool:
        return any(
            is_relative_to(path, p) for p in self.__config.compiler.solc.ignore_paths
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Optional, Tuple, Union

from intervaltree import IntervalTree

from ...types import Contract, Enum, Struct, UserDefinedValueType

if TYPE_CHECKING:
    from ..declaration.variable_declaration import VariableDeclaration
    from ..expression.new_expression import NewExpression
    from ..meta.inheritance_specifier import InheritanceSpecifier
    from ..meta.override_specifier import OverrideSpecifier
    from ..meta.using_for_directive import UsingForDirective
    from ..type_name.array_type_name import ArrayTypeName
    from ..type_name.mapping import Mapping

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.abc import DeclarationAbc
from woke.ast.ir.meta.identifier_path import (
    IDENTIFIER_RE,
    IdentifierPath,
    IdentifierPathPart,
)
from woke.ast.ir.type_name.abc import TypeNameAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcUserDefinedTypeName


class UserDefinedTypeName(TypeNameAbc):
    """
    User defined type name represents a name path to a user defined type. Path parts are separated by dots.
    In Solidity 0.8.0 a new IR node ([IdentifierPath][woke.ast.ir.meta.identifier_path.IdentifierPath]) was introduced to replace [UserDefinedTypeName][woke.ast.ir.type_name.user_defined_type_name.UserDefinedTypeName] in some cases.

    !!! example
        A user defined type name can be used:

        - inside a [VariableDeclaration][woke.ast.ir.declaration.variable_declaration.VariableDeclaration]:
            - `:::solidity Interface.Struct` in line 18,
            - `:::solidity Interface.Enum` in line 26,
        - inside a [NewExpression][woke.ast.ir.expression.new_expression.NewExpression]:
            - `:::solidity Contract` in line 20,
        - inside an [InheritanceSpecifier][woke.ast.ir.meta.inheritance_specifier.InheritanceSpecifier]:
            - `:::solidity Interface` in line 23,
        - inside an [OverrideSpecifier][woke.ast.ir.meta.override_specifier.OverrideSpecifier]:
            - `:::solidity Interface` in line 30,
        - inside a [UsingForDirective][woke.ast.ir.meta.using_for_directive.UsingForDirective]:
            - `:::solidity Lib` in line 24,
            - `:::solidity Interface.Struct` in line 24,
        - inside an [ArrayTypeName][woke.ast.ir.type_name.array_type_name.ArrayTypeName]:
            - `:::solidity Interface.Enum` in line 27,
        - inside a [Mapping][woke.ast.ir.type_name.mapping.Mapping]:
            - both occurrences of `:::solidity Interface.Enum` in line 28.

        ```solidity linenums="1"
        pragma solidity 0.7;

        interface Interface {
            enum Enum {
                READY,
                WAITING
            }

            struct Struct {
                uint a;
            }

            function foo() external;
        }

        library Lib {}

        function tmp(Interface.Struct memory s) {
            s.a = 5;
            new Contract();
        }

        contract Contract is Interface {
            using Lib for Interface.Struct;

            Interface.Enum state;
            Interface.Enum[] states;
            mapping(Interface.Enum => Interface.Enum) map;

            function foo() external override(Interface) {
            }
        }
        ```
    """

    _ast_node: SolcUserDefinedTypeName
    _parent: Union[
        VariableDeclaration,
        NewExpression,
        InheritanceSpecifier,
        OverrideSpecifier,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]

    _referenced_declaration_id: AstNodeId
    _contract_scope_id: Optional[AstNodeId]
    _name: Optional[str]
    _path_node: Optional[IdentifierPath]
    _parts: Optional[IntervalTree]

    def __init__(
        self,
        init: IrInitTuple,
        user_defined_type_name: SolcUserDefinedTypeName,
        parent: SolidityAbc,
    ):
        super().__init__(init, user_defined_type_name, parent)
        self._name = user_defined_type_name.name
        self._referenced_declaration_id = user_defined_type_name.referenced_declaration
        assert self._referenced_declaration_id >= 0
        self._contract_scope_id = user_defined_type_name.contract_scope
        if user_defined_type_name.path_node is None:
            self._path_node = None
            matches = list(IDENTIFIER_RE.finditer(self._source))
            groups_count = len(matches)
            assert groups_count > 0

            self._parts = IntervalTree()
            for i, match in enumerate(matches):
                name = match.group(0).decode("utf-8")
                start = self.byte_location[0] + match.start()
                end = self.byte_location[0] + match.end()
                self._parts[start:end] = IdentifierPathPart(
                    self,
                    init,
                    (start, end),
                    name,
                    self._referenced_declaration_id,
                    groups_count - i - 1,
                )
        else:
            self._path_node = IdentifierPath(
                init, user_defined_type_name.path_node, self
            )
            self._parts = None

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._path_node is not None:
            yield from self._path_node

    @property
    def parent(
        self,
    ) -> Union[
        VariableDeclaration,
        NewExpression,
        InheritanceSpecifier,
        OverrideSpecifier,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def type(self) -> Union[Contract, Struct, Enum, UserDefinedValueType]:
        """
        Returns:
            Type description.
        """
        t = super().type
        assert isinstance(t, (Contract, Struct, Enum, UserDefinedValueType))
        return t

    @property
    def name(self) -> str:
        """
        !!! note
            Should be the same as [source][woke.ast.ir.abc.IrAbc.source] and is the same as [path_node.name][woke.ast.ir.meta.identifier_path.IdentifierPath.name] if [path_node][woke.ast.ir.type_name.user_defined_type_name.UserDefinedTypeName.path_node] is not `None`.
        Returns:
            Name of the user defined type as it appears in the source code.
        """
        if self._name is None:
            assert self._path_node is not None
            self._name = self._path_node.name
        return self._name

    @property
    def identifier_path_parts(self) -> Tuple[IdentifierPathPart, ...]:
        """
        Returns:
            Parts of the user defined type name.
        """
        if self._path_node is not None:
            return self._path_node.identifier_path_parts

        assert self._parts is not None
        return tuple(interval.data for interval in sorted(self._parts))

    def identifier_path_part_at(self, byte_offset: int) -> Optional[IdentifierPathPart]:
        """
        Args:
            byte_offset: Byte offset in the source file.
        Returns:
            Identifier path part at the given byte offset, if any.
        """
        if self._path_node is not None:
            return self._path_node.identifier_path_part_at(byte_offset)

        assert self._parts is not None
        intervals = self._parts.at(byte_offset)
        assert len(intervals) <= 1
        if len(intervals) == 0:
            return None
        return intervals.pop().data

    @property
    def referenced_declaration(self) -> DeclarationAbc:
        """
        Returns:
            Declaration IR node referenced by this user defined type name.
        """
        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, DeclarationAbc)
        return node

    @property
    def path_node(self) -> Optional[IdentifierPath]:
        """
        Always present since Solidity 0.8.0. If not `None`, it represents the same source code as this node ([byte_location][woke.ast.ir.abc.IrAbc.byte_location] properties are the same) and references the same declaration.
        Returns:
            Identifier path IR node.
        """
        return self._path_node



from .api import (
    DetectorAbc,
    DetectorResult,
    detect,
    detector,
    print_detection,
    print_detectors,
)
from .balance_state_var import UnsafeAddressBalanceUseDetector
from .call_options_not_called import FunctionCallOptionsNotCalledDetector
from .overflow_calldata_tuple_reencoding_bug import (
    OverflowCalldataTupleReencodingBugDetector,
)
from .reentrancy import ReentrancyDetector
from .unchecked_return_value import UncheckedFunctionReturnValueDetector
from .unsafe_delegatecall import UnsafeDelegatecallDetector
from .unsafe_selfdestruct import UnsafeSelfdestructDetector

from . import axelar  # isort:skip



from functools import lru_cache
from typing import Iterator, Set, Tuple

from woke.ast.enums import ModifiesStateFlag, UnaryOpOperator
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcUnaryOperation


class UnaryOperation(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcUnaryOperation
    _parent: SolidityAbc

    _operator: UnaryOpOperator
    _prefix: bool
    _sub_expression: ExpressionAbc

    def __init__(
        self,
        init: IrInitTuple,
        unary_operation: SolcUnaryOperation,
        parent: SolidityAbc,
    ):
        super().__init__(init, unary_operation, parent)
        self._operator = unary_operation.operator
        self._prefix = unary_operation.prefix
        self._sub_expression = ExpressionAbc.from_ast(
            init, unary_operation.sub_expression, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._sub_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def operator(self) -> UnaryOpOperator:
        return self._operator

    @property
    def prefix(self) -> bool:
        return self._prefix

    @property
    def sub_expression(self) -> ExpressionAbc:
        return self._sub_expression

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.sub_expression.modifies_state

        if (
            self.operator
            in {
                UnaryOpOperator.PLUS_PLUS,
                UnaryOpOperator.MINUS_MINUS,
                UnaryOpOperator.DELETE,
            }
            and self.sub_expression.is_ref_to_state_variable
        ):
            ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
        return ret



import logging
import random
from datetime import datetime, timedelta
from typing import Callable, Counter, List, Optional, Tuple

import brownie
import IPython

from .utils import partition

Methods = List[Tuple[Callable, str]]


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Campaign:
    __sequence_constructor: Callable

    def __init__(self, sequence_constructor: Callable) -> None:
        self.__sequence_constructor = sequence_constructor

    def run(
        self,
        sequences_count: int,
        flows_count: int,
        run_for_seconds: Optional[int] = None,
        dry_run: bool = False,
    ):
        init_timestamp = datetime.now()
        brownie.chain.snapshot()

        for i in range(sequences_count):
            if (
                run_for_seconds is not None
                and datetime.now()
                >= init_timestamp + timedelta(seconds=run_for_seconds)
            ):
                break
            logger.info(self.__format_heading(f"SEQUENCE {i}"))
            brownie.chain.revert()
            seq = self.__sequence_constructor()

            flows, _ = self.__get_methods(seq, attr="flow")
            invs, _ = self.__get_methods(seq, attr="invariant")

            called = Counter[str]()
            # point_coverage = Counter[str]()

            for j in range(flows_count):
                if flows:

                    def meets_preconditions(flow):
                        flow = flow[0]
                        meets_precondition = not hasattr(
                            flow, "precondition"
                        ) or flow.precondition(seq)
                        meets_max_times = (
                            not hasattr(flow, "max_times")
                            or called[flow] < flow.max_times
                        )
                        return meets_precondition and meets_max_times

                    flows_p, flows_not_p = partition(flows, meets_preconditions)
                    if flows_not_p:
                        logger.info(f"\nThe following flows' preconditions are falsy:")
                        for flow_not_p in flows_not_p:
                            logger.info(f"    {flow_not_p[1]}")
                    else:
                        logger.info(f"\n")

                    if len(flows_p) == 0:
                        logger.warning(
                            "There are no rules satisfying conditions. Terminating the sequence..."
                        )
                        break

                    flows_weights = []

                    for flow in flows_p:
                        if hasattr(flow[0], "weight"):
                            flows_weights.append(flow[0].weight)
                        else:
                            flows_weights.append(100)

                    logger.info("Weights:")
                    for idx in range(len(flows_p)):
                        logger.info(f"    {flows_p[idx][1]}: {flows_weights[idx]}")
                    flow = random.choices(flows_p, weights=flows_weights, k=1)[0]
                    called.update((flow[0],))  # type: ignore

                    logger.info(
                        f'\n{f"FLOW...":<9} {j:>4} IN SEQUENCE {i:>5} {flow[1]}:'
                    )
                    flow[0]()

                if not dry_run and invs:
                    for idx, inv in enumerate(invs):
                        logger.info(f'{"inv...":<33}{inv[1]}')
                        inv[0]()
                        del inv

            del invs, flows
            # point_coverage += seq.point_coverage
            # logger.info(self.__format_heading("Sequence point coverage:"))
            # self.__log_point_coverage(seq.point_coverage)
            # logger.info(self.__format_heading("Campaign point coverage:"))
            # self.__log_point_coverage(point_coverage)
            del seq
        logger.info(f"\nRan {flows_count} flows. All flows and invariants passed.")

    @staticmethod
    def __is_ign(m: Callable) -> bool:
        return hasattr(m, "ignore") and m.ignore

    @staticmethod
    def __get_methods(o, attr: str) -> Tuple[Methods, Methods]:
        """get not ignored and ignored methods of object {o} beginning with {prefix} and having attribute a truthy attribute {attr}"""
        ms = []
        for m_str in dir(o):
            m = getattr(o, m_str)
            if hasattr(m, attr) and getattr(m, attr):
                # m_str_body = m_str.split(prefix)[1] if prefix else m_str
                # if m_str_body.startswith('_'):
                #     m_str_body = m_str_body[1:]
                ms.append((m, m_str))

        # invariant: at this point, ms contains all relevant methods
        # now find those that aren't and are ignored

        ms_ign, ms_not_ign = partition(ms, lambda m: Campaign.__is_ign(m[0]))

        if ms_ign:
            s = "Ignoring:\n" + "\n".join(m[1] for m in ms_ign)
            logger.info(s)
        del ms
        return ms_not_ign, ms_ign

    @staticmethod
    def __format_heading(s: str) -> str:
        res = ""
        res += "\n"
        res += "-" * len(s) + "\n"
        res += s
        res += "\n"
        res += "-" * len(s) + "\n"
        res += "\n"
        return res

    @staticmethod
    def __log_point_coverage(c: Counter):
        items = list(c.items())
        items.sort()
        for key, count in items:
            print(f"{key:.<80}{count:.>4}")



from pathlib import Path

import pydantic
import pytest

from woke.config import WokeConfig
from woke.config.data_model import SolcRemapping
from woke.core.enums import EvmVersionEnum
from woke.core.solidity_version import SolidityVersion

sources_path = (Path(__file__).parent / "config_sources").resolve()


@pytest.mark.platform_dependent
def test_config_empty():
    config = WokeConfig(woke_root_path=sources_path)
    assert config.woke_root_path.resolve() == sources_path.resolve()
    assert len(config.loaded_files) == 0

    file_path = (sources_path / "empty.toml").resolve()
    config.load(file_path)
    assert len(config.loaded_files) == 1
    assert file_path in config.loaded_files
    assert len(config.compiler.solc.remappings) == 0


@pytest.mark.platform_dependent
def test_config_simple():
    config = WokeConfig(woke_root_path=sources_path)
    config.load_configs()  # should not have any effect
    assert len(config.loaded_files) == 0
    assert config.woke_root_path.resolve() == sources_path.resolve()
    assert len(config.compiler.solc.remappings) == 0

    config.load(sources_path / "a.toml")
    assert config.woke_root_path.resolve() == sources_path.resolve()
    assert len(config.loaded_files) == 4
    assert (sources_path / "a.toml").resolve() in config.loaded_files
    assert (sources_path / "b.toml").resolve() in config.loaded_files
    assert (sources_path / "c.toml").resolve() in config.loaded_files
    assert (sources_path / "empty.toml").resolve() in config.loaded_files
    assert len(config.compiler.solc.remappings) == 1
    assert config.compiler.solc.remappings[0] == SolcRemapping(
        context=None, prefix="xyz", target=None
    )
    assert str(config.compiler.solc.remappings[0]) == ":xyz="
    assert len(config.compiler.solc.include_paths) == 1
    assert sources_path in config.compiler.solc.include_paths
    assert len(config.compiler.solc.allow_paths) == 1
    assert (sources_path / "../").resolve() in config.compiler.solc.allow_paths


def test_config_from_dict():
    config_dict = {
        "compiler": {
            "solc": {
                "allow_paths": ["."],
                "evm_version": "london",
                "include_paths": ["../"],
                "remappings": ["hardhat/=node_modules/hardhat/"],
                "target_version": "0.8.12",
            }
        }
    }
    config = WokeConfig.fromdict(
        config_dict,
        woke_root_path=sources_path,
        project_root_path=(sources_path / "containing_global_conf"),
    )

    assert len(config.compiler.solc.allow_paths) == 1
    assert sources_path / "containing_global_conf" in config.compiler.solc.allow_paths
    assert config.compiler.solc.evm_version == EvmVersionEnum.LONDON
    assert len(config.compiler.solc.include_paths) == 1
    assert sources_path in config.compiler.solc.include_paths
    assert len(config.compiler.solc.remappings) == 1
    assert config.compiler.solc.remappings[0] == SolcRemapping(
        context=None, prefix="hardhat/", target="node_modules/hardhat/"
    )
    assert config.compiler.solc.target_version == SolidityVersion.fromstring("0.8.12")


@pytest.mark.platform_dependent
def test_config_global():
    config = WokeConfig(woke_root_path=(sources_path / "containing_global_conf"))
    assert len(config.loaded_files) == 0
    config.load_configs()
    assert len(config.loaded_files) == 2
    assert (
        sources_path / "containing_global_conf" / "config.toml"
    ).resolve() in config.loaded_files
    assert (sources_path / "empty.toml").resolve() in config.loaded_files
    assert len(config.compiler.solc.remappings) == 2
    assert config.compiler.solc.remappings[0] == SolcRemapping(
        context=None, prefix="https://url.com", target=None
    )
    assert config.compiler.solc.remappings[1] == SolcRemapping(
        context=None, prefix="123", target="xyz"
    )


@pytest.mark.platform_dependent
def test_config_project():
    config = WokeConfig(
        project_root_path=(sources_path / "containing_project_conf"),
        woke_root_path=(sources_path / "containing_global_conf"),
    )
    config.load_configs()

    assert len(config.compiler.solc.remappings) == 1
    assert config.compiler.solc.remappings[0] == SolcRemapping(
        context=None, prefix="woke", target="test-target"
    )


@pytest.mark.platform_dependent
def test_config_project_path_not_dir():
    with pytest.raises(ValueError):
        config = WokeConfig(
            project_root_path=(sources_path / "a,toml"), woke_root_path=sources_path
        )


@pytest.mark.platform_dependent
def test_config_import_abs_path():
    tmp_file = sources_path / "tmp_bHtvhGrDp6.toml"
    abs_path = (sources_path / "a.toml").resolve()
    content = 'subconfigs = ["{path}"]'.format(path=str(abs_path).replace("\\", "\\\\"))

    tmp_file.write_text(content)
    config = WokeConfig(woke_root_path=sources_path)
    # this one should load: a.toml -> b.toml -> c.toml
    config.load(tmp_file)

    assert len(config.loaded_files) == 5
    assert tmp_file.resolve() in config.loaded_files
    assert (sources_path / "a.toml").resolve() in config.loaded_files
    assert (sources_path / "b.toml").resolve() in config.loaded_files
    assert (sources_path / "c.toml").resolve() in config.loaded_files
    assert (sources_path / "empty.toml").resolve() in config.loaded_files
    assert len(config.compiler.solc.remappings) == 1
    assert config.compiler.solc.remappings[0] == SolcRemapping(
        context=None, prefix="xyz", target=None
    )
    assert len(config.compiler.solc.include_paths) == 1
    assert sources_path in config.compiler.solc.include_paths

    tmp_file.unlink()


@pytest.mark.platform_dependent
def test_config_invalid_format():
    config = WokeConfig(woke_root_path=sources_path)

    with pytest.raises(pydantic.ValidationError):
        config.load(sources_path / "invalid_1.toml")
    with pytest.raises(pydantic.ValidationError):
        config.load(sources_path / "invalid_2.toml")
    with pytest.raises(pydantic.ValidationError):
        config.load(sources_path / "invalid_3.toml")


@pytest.mark.platform_dependent
def test_config_cyclic_import():
    config = WokeConfig(woke_root_path=sources_path)
    with pytest.raises(ValueError):
        config.load(sources_path / "cyclic_x.toml")

    with pytest.raises(ValueError):
        config.load(sources_path / "cyclic_1.toml")

    with pytest.raises(ValueError):
        config.load(sources_path / "cyclic_2.toml")



from functools import lru_cache
from typing import Iterator, Set, Tuple

from woke.ast.enums import BinaryOpOperator, ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcBinaryOperation

from .abc import ExpressionAbc


class BinaryOperation(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcBinaryOperation
    _parent: SolidityAbc  # TODO: make this more specific

    _left_expression: ExpressionAbc
    _operator: BinaryOpOperator
    _right_expression: ExpressionAbc

    def __init__(
        self,
        init: IrInitTuple,
        binary_operation: SolcBinaryOperation,
        parent: SolidityAbc,
    ):
        super().__init__(init, binary_operation, parent)
        self._operator = binary_operation.operator
        self._left_expression = ExpressionAbc.from_ast(
            init, binary_operation.left_expression, self
        )
        self._right_expression = ExpressionAbc.from_ast(
            init, binary_operation.right_expression, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._left_expression
        yield from self._right_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def operator(self) -> BinaryOpOperator:
        return self._operator

    @property
    def left_expression(self) -> ExpressionAbc:
        return self._left_expression

    @property
    def right_expression(self) -> ExpressionAbc:
        return self._right_expression

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return (
            self.left_expression.modifies_state | self.right_expression.modifies_state
        )



import pathlib

import click
from click.core import Context

from woke.config import WokeConfig
from woke.fuzzer.abi_to_type import generate_types
from woke.utils import file_utils


@click.group(name="init")
@click.pass_context
def run_init(ctx: Context):
    """Create default project skeleton."""
    config = WokeConfig(woke_root_path=ctx.obj["woke_root_path"])
    config.load_configs()
    ctx.obj["config"] = config


@run_init.command(name="types")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite existing types.",
)
@click.pass_context
def init_types(ctx: Context, force: bool) -> None:
    """Generate Python contract types from Solidity ABI."""
    config: WokeConfig = ctx.obj["config"]
    generate_types(config, force)


@run_init.command(name="fuzz")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite existing pytypes and tests directories",
)
@click.pass_context
def init_fuzz(ctx: Context, force: bool) -> None:
    """Generate Python contract types and create example fuzz tests."""
    config: WokeConfig = ctx.obj["config"]

    generate_types(config, force)

    examples_dir = pathlib.Path(__file__).parent.parent.resolve() / "examples/fuzzer"
    tests_dir = config.project_root_path / "tests"
    file_utils.copy_dir(examples_dir, tests_dir, overwrite=force)



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

import woke.ast.ir.yul as yul

from ...ast.enums import GlobalSymbolsEnum
from ...ast.ir.abc import IrAbc
from ...ast.ir.declaration.abc import DeclarationAbc
from ...ast.ir.declaration.contract_definition import ContractDefinition
from ...ast.ir.expression.identifier import Identifier
from ...ast.ir.expression.member_access import MemberAccess
from ...ast.ir.meta.identifier_path import IdentifierPath
from ...ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from ...core.solidity_version import SemanticVersion
from ..common_structures import (
    MarkupContent,
    MarkupKind,
    MessageType,
    Position,
    Range,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from ..context import LspContext
from ..lsp_data_model import LspModel
from ..utils import position_within_range, uri_to_path
from ..utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class HoverClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    content_format: Optional[List[MarkupKind]]


class HoverOptions(WorkDoneProgressOptions):
    pass


class HoverRegistrationOptions(TextDocumentRegistrationOptions, HoverOptions):
    pass


class HoverParams(TextDocumentPositionParams, WorkDoneProgressParams):
    pass


class MarkedString(LspModel):
    language: str
    value: str


MarkedStringType = Union[str, MarkedString]


class Hover(LspModel):
    contents: Union[MarkedStringType, List[MarkedStringType], MarkupContent]
    range: Optional[Range]


def _append_openzeppelin_docs(
    node: ContractDefinition, version: SemanticVersion
) -> str:
    source_unit = node.parent
    version_string = f"{version.major}.x"
    url_base = f"https://docs.openzeppelin.com/contracts/{version_string}"
    doc_url = None
    api_doc_url = None

    if "openzeppelin/contracts/access" in source_unit.source_unit_name:
        doc_url = "access-control"
        api_doc_url = "access"
    elif "openzeppelin/contracts/crosschain" in source_unit.source_unit_name:
        doc_url = "crosschain"
        api_doc_url = "crosschain"
    elif "openzeppelin/contracts/finance" in source_unit.source_unit_name:
        if node.name == "PaymentSplitter":
            doc_url = "utilities#payment"
        api_doc_url = "finance"
    elif "openzeppelin/contracts/governance" in source_unit.source_unit_name:
        doc_url = "governance"
        api_doc_url = "governance"
    elif (
        "openzeppelin/contracts/interfaces" in source_unit.source_unit_name
        or "openzeppelin/contracts/token" in source_unit.source_unit_name
    ):
        if "ERC20" in node.name:
            doc_url = "erc20"
            api_doc_url = "token/erc20"
        elif "ERC721" in node.name:
            doc_url = "erc721"
            api_doc_url = "token/erc721"
        elif "ERC777" in node.name:
            doc_url = "erc777"
            api_doc_url = "token/erc777"
        elif "ERC1155" in node.name:
            doc_url = "erc1155"
            api_doc_url = "token/erc1155"
        elif node.name.startswith("I"):
            api_doc_url = "interfaces"
        else:
            api_doc_url = "token/common"
    elif "openzeppelin/contracts/metatx" in source_unit.source_unit_name:
        api_doc_url = "metatx"
    elif "openzeppelin/contracts/proxy" in source_unit.source_unit_name:
        api_doc_url = "proxy"
    elif "openzeppelin/contracts/security" in source_unit.source_unit_name:
        if node.name == "PullPayment":
            doc_url = "utilities#payment"
        api_doc_url = "security"
    elif "openzeppelin/contracts/utils" in source_unit.source_unit_name:
        doc_url = "utilities"
        api_doc_url = "utils"
    elif "openzeppelin/contracts/GSN" in source_unit.source_unit_name:
        doc_url = "gsn"
        api_doc_url = "gsn"
    elif "openzeppelin/contracts/cryptography" in source_unit.source_unit_name:
        doc_url = "utilities#cryptography"
        api_doc_url = "cryptography"
    elif "openzeppelin/contracts/drafts" in source_unit.source_unit_name:
        api_doc_url = "drafts"
    elif "openzeppelin/contracts/math" in source_unit.source_unit_name:
        doc_url = "utilities#math"
        api_doc_url = "math"
    elif "openzeppelin/contracts/payment" in source_unit.source_unit_name:
        doc_url = "utilities#payment"
        api_doc_url = "payment"
    elif "openzeppelin/contracts/presets" in source_unit.source_unit_name:
        if "ERC20" in node.name:
            doc_url = "erc20#Presets"
        elif "ERC721" in node.name:
            doc_url = "erc721#Presets"
        elif "ERC777" in node.name:
            doc_url = "erc777#Presets"
        elif "ERC1155" in node.name:
            doc_url = "erc1155#Presets"
        api_doc_url = "presets"

    ret = ""
    if doc_url is not None:
        ret += f"\n\n[OpenZeppelin documentation]({url_base}/{doc_url})"
    if api_doc_url is not None:
        ret += f"\n\n[OpenZeppelin API documentation]({url_base}/api/{api_doc_url}#{node.name})"
    return ret


def _get_results_from_node(
    original_node: IrAbc,
    position: Position,
    byte_offset: int,
    context: LspContext,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[Tuple[str, Tuple[int, int]]]:
    original_node_location: Tuple[int, int] = original_node.byte_location
    logger.debug(f"Found node {original_node}")

    if isinstance(original_node, DeclarationAbc):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if not position_within_range(position, name_location_range):
            return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
        original_node_location = part.byte_location
    elif isinstance(original_node, yul.Identifier):
        external_reference = original_node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration
        original_node_location = external_reference.byte_location
    else:
        node = original_node

    if not isinstance(node, DeclarationAbc):
        return None

    value = "```solidity\n" + node.declaration_string + "\n```"
    if (
        isinstance(node, ContractDefinition)
        and context.openzeppelin_contracts_version is not None
        and context.openzeppelin_contracts_version >= "2.0.0"
    ):
        value += _append_openzeppelin_docs(node, context.openzeppelin_contracts_version)

    return value, original_node_location


async def _get_hover_from_cache(path: Path, position: Position, context: LspContext):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        location = node.name_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, old_byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, location = result
    forward_changes = context.compiler.get_last_compilation_forward_changes(path)
    if forward_changes is None:
        raise Exception("No forward changes found")
    new_start = changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
    new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(path, (new_start, new_end)),
    )


async def hover(context: LspContext, params: HoverParams) -> Optional[Hover]:
    logger.debug(
        f"Hover for file {params.text_document.uri} at position {params.position} requested"
    )
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.interval_trees:
        try:
            return await _get_hover_from_cache(path, params.position, context)
        except Exception:
            return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(f"Found {len(nodes)} nodes at position {params.position}")
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, original_node_location = result
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(
            path, original_node_location
        ),
    )



import importlib.resources
import importlib.util
import inspect
import multiprocessing
import platform
import sys
import time
from pathlib import Path
from typing import Callable, Iterable, Tuple

import click

from woke.config import WokeConfig
from woke.fuzzer import fuzz

from .console import console


def _get_module_name(path: Path, root: Path) -> str:
    path = path.with_suffix("")
    return ".".join(path.relative_to(root).parts)


@click.command(name="fuzz")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--process-count",
    "-n",
    default=(multiprocessing.cpu_count()),
    help="Number of processes to create for fuzzing.",
)
@click.option("--seed", "-s", "seeds", multiple=True, type=str, help="Random seeds")
@click.option(
    "--passive",
    is_flag=True,
    default=False,
    help="Print one process output into console, run other in background.",
)
@click.option(
    "--network",
    type=str,
    default="development",
    help="Choose brownie dev chain. Default is 'development' for ganache",
)
@click.pass_context
def run_fuzz(
    ctx: click.Context,
    paths: Tuple[str],
    process_count: int,
    seeds: Tuple[str],
    passive: bool,
    network: str,
) -> None:
    """Run Woke fuzzer."""
    config = WokeConfig(woke_root_path=ctx.obj["woke_root_path"])
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    random_seeds = [bytes.fromhex(seed) for seed in seeds]
    if len(paths) == 0:
        paths = (str(config.project_root_path / "tests"),)

    py_files = set()

    for path in paths:
        fuzz_path = Path(path).resolve()

        if fuzz_path.is_file() and fuzz_path.match("*.py"):
            py_files.add(fuzz_path)
        elif fuzz_path.is_dir():
            for p in fuzz_path.rglob("test_*.py"):
                if p.is_file():
                    py_files.add(p)
        else:
            raise ValueError(f"'{fuzz_path}' is not a Python file or directory.")

    fuzz_functions = []

    sys.path.insert(0, str(config.project_root_path))
    for file in py_files:
        module_name = _get_module_name(file, config.project_root_path)
        module_spec = importlib.util.spec_from_file_location(module_name, file)
        if module_spec is None or module_spec.loader is None:
            raise ValueError()
        module = importlib.util.module_from_spec(module_spec)
        sys.modules[module_name] = module
        module_spec.loader.exec_module(module)

        functions: Iterable[Callable] = (
            func
            for _, func in inspect.getmembers(module, inspect.isfunction)
            if func.__module__ == module_name and func.__name__.startswith("test")
        )
        for func in functions:
            console.print(
                f"Found '{func.__name__}' function in '{func.__module__}' file."
            )
            fuzz_functions.append(func)

    logs_dir = config.project_root_path / ".woke-logs" / "fuzz" / str(int(time.time()))
    logs_dir.mkdir(parents=True, exist_ok=False)
    latest_dir = logs_dir.parent / "latest"

    # create `latest` symlink
    if platform.system() != "Windows":
        if latest_dir.is_symlink():
            latest_dir.unlink()
        latest_dir.symlink_to(logs_dir, target_is_directory=True)

    for func in fuzz_functions:
        console.print("\n\n")
        console.print(f"Fuzzing '{func.__name__}' in '{func.__module__}'.")
        fuzz(config, func, process_count, random_seeds, logs_dir, passive, network)



from __future__ import annotations

from typing import TYPE_CHECKING

from .abc import YulAbc

if TYPE_CHECKING:
    from .block import Block


class Break(YulAbc):
    """
    TBD
    """
    _parent: Block

    @property
    def parent(self) -> Block:
        return self._parent



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.type_name.abc import TypeNameAbc
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.ast.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from woke.ast.ir.declaration.contract_definition import ContractDefinition
    from .source_unit import SourceUnit

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.nodes import (
    SolcIdentifierPath,
    SolcUserDefinedTypeName,
    SolcUsingForDirective,
)


class UsingForDirective(SolidityAbc):
    """
    !!! note
        Either [functions][woke.ast.ir.meta.using_for_directive.UsingForDirective.functions] or [library_name][woke.ast.ir.meta.using_for_directive.UsingForDirective.library_name] must be set.
    !!! example
        Lines 13 and 14 in the following example:
        ```solidity linenums="1"
        function add(uint a, uint b) pure returns (uint) {
            return a + b;
        }

        library SafeMath {
            function sub(uint a, uint b) pure returns (uint) {
                require(b <= a);
                return a - b;
            }
        }

        contract C {
            using SafeMath for uint;
            using {add as add2} for uint;
        }
        ```
    """
    _ast_node: SolcUsingForDirective
    _parent: Union[ContractDefinition, SourceUnit]

    _functions: Optional[List[IdentifierPath]]
    _library_name: Optional[Union[IdentifierPath, UserDefinedTypeName]]
    _type_name: Optional[TypeNameAbc]

    def __init__(
        self,
        init: IrInitTuple,
        using_for_directive: SolcUsingForDirective,
        parent: Union[ContractDefinition, SourceUnit],
    ):
        super().__init__(init, using_for_directive, parent)

        if using_for_directive.function_list is None:
            self._functions = None
        else:
            self._functions = [
                IdentifierPath(init, function.function, self)
                for function in using_for_directive.function_list
            ]

        if using_for_directive.library_name is None:
            self._library_name = None
        elif isinstance(using_for_directive.library_name, SolcUserDefinedTypeName):
            self._library_name = UserDefinedTypeName(
                init, using_for_directive.library_name, self
            )
        elif isinstance(using_for_directive.library_name, SolcIdentifierPath):
            self._library_name = IdentifierPath(
                init, using_for_directive.library_name, self
            )

        if using_for_directive.type_name is None:
            self._type_name = None
        else:
            self._type_name = TypeNameAbc.from_ast(
                init, using_for_directive.type_name, self
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._functions is not None:
            for function in self._functions:
                yield from function
        if self._library_name is not None:
            yield from self._library_name
        if self._type_name is not None:
            yield from self._type_name

    @property
    def parent(self) -> Union[ContractDefinition, SourceUnit]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def functions(self) -> Optional[Tuple[IdentifierPath]]:
        """
        Is only set in the case of `:::solidity using {function1, function2} for TypeName;` directive type.
        Returns:
            List of functions that are bound to the target type.
        """
        if self._functions is None:
            return None
        return tuple(self._functions)

    @property
    def library_name(self) -> Optional[Union[IdentifierPath, UserDefinedTypeName]]:
        """
        Is only set in the case of `:::solidity using LibraryName for TypeName;` directive type.
        Returns:
            IR node referencing the library ([ContractDefinition][woke.ast.ir.declaration.contract_definition.ContractDefinition] of the [ContractKind.LIBRARY][woke.ast.enums.ContractKind.LIBRARY] kind) that is bound to the target type.
        """
        return self._library_name

    @property
    def type_name(self) -> Optional[TypeNameAbc]:
        """
        Is `None` in the case of `:::solidity using Lib for *;`.
        Returns:
            Type name that is bound to the functions or library.
        """
        return self._type_name



import asyncio
import logging
import platform
import shutil
import time
from collections import deque
from itertools import chain
from json import JSONDecodeError
from pathlib import Path, PurePath
from typing import Collection, Dict, Iterable, List, Mapping, Optional, Set, Tuple

import aiofiles
import networkx as nx
from pathvalidate import sanitize_filename  # type: ignore
from pydantic import ValidationError
from rich.progress import Progress

from woke.config import WokeConfig
from woke.core.solidity_version import (
    SolidityVersion,
    SolidityVersionRange,
    SolidityVersionRanges,
)
from woke.regex_parsing import SoliditySourceParser
from woke.svm import SolcVersionManager

from .build_data_model import CompilationUnitBuildInfo, ProjectBuildInfo
from .compilation_unit import CompilationUnit
from .exceptions import CompilationError, CompilationResolveError
from .solc_frontend import (
    SolcFrontend,
    SolcInputOptimizerSettings,
    SolcInputSettings,
    SolcOutput,
    SolcOutputContractInfo,
    SolcOutputSelectionEnum,
    SolcOutputSourceInfo,
)
from .source_path_resolver import SourcePathResolver
from .source_unit_name_resolver import SourceUnitNameResolver

logger = logging.getLogger(__name__)


class SolidityCompiler:
    __config: WokeConfig
    __svm: SolcVersionManager
    __solc_frontend: SolcFrontend
    __source_unit_name_resolver: SourceUnitNameResolver
    __source_path_resolver: SourcePathResolver

    def __init__(self, woke_config: WokeConfig):
        self.__config = woke_config
        self.__svm = SolcVersionManager(woke_config)
        self.__solc_frontend = SolcFrontend(woke_config)
        self.__source_unit_name_resolver = SourceUnitNameResolver(woke_config)
        self.__source_path_resolver = SourcePathResolver(woke_config)

    def build_graph(
        self,
        files: Collection[Path],
        modified_files: Mapping[Path, str],
        ignore_errors: bool = False,
    ) -> nx.DiGraph:
        # source unit name, full path, file content
        source_units_queue: deque[Tuple[PurePath, Path, Optional[str]]] = deque()
        source_units: Dict[PurePath, Path] = {}

        # for every source file resolve a source unit name
        for file in files:
            try:
                file = file.resolve(strict=True)
            except FileNotFoundError:
                if file in modified_files:
                    pass
                elif ignore_errors:
                    continue
                else:
                    raise

            source_unit_name = self.__source_unit_name_resolver.resolve_cmdline_arg(
                str(file)
            )
            if source_unit_name in source_units:
                first = str(source_units[source_unit_name])
                second = str(file)
                raise CompilationError(
                    f"Same source unit name `{source_unit_name}` for multiple source files:\n{first}\n{second}"
                )
            source_units[source_unit_name] = file
            content = modified_files.get(file, None)
            source_units_queue.append((source_unit_name, file, content))

        graph = nx.DiGraph()

        # recursively process all sources
        while len(source_units_queue) > 0:
            source_unit_name, path, content = source_units_queue.pop()
            if content is None:
                try:
                    versions, imports, h, content = SoliditySourceParser.parse(
                        path, ignore_errors
                    )
                except UnicodeDecodeError:
                    continue
            else:
                versions, imports, h = SoliditySourceParser.parse_source(
                    content, ignore_errors
                )
            graph.add_node(
                source_unit_name,
                path=path,
                versions=versions,
                hash=h,
                content=content,
                unresolved_imports=set(),
            )
            source_units[source_unit_name] = path

            for _import in imports:
                import_unit_name = self.__source_unit_name_resolver.resolve_import(
                    source_unit_name, _import
                )
                try:
                    import_path = self.__source_path_resolver.resolve(
                        import_unit_name, modified_files.keys()
                    ).resolve()
                except (FileNotFoundError, CompilationResolveError):
                    if ignore_errors:
                        graph.nodes[source_unit_name]["unresolved_imports"].add(
                            import_unit_name
                        )
                        continue
                    raise

                if import_unit_name in source_units:
                    other_path = source_units[import_unit_name]
                    if import_path != other_path:
                        raise ValueError(
                            f"Same source unit name `{import_unit_name}` for multiple source files:\n{import_path}\n{other_path}"
                        )
                else:
                    source_units_queue.append(
                        (
                            import_unit_name,
                            import_path,
                            modified_files.get(import_path, None),
                        )
                    )
                graph.add_edge(import_unit_name, source_unit_name)
        return graph

    @staticmethod
    def build_compilation_units_maximize(graph: nx.DiGraph) -> List[CompilationUnit]:
        """
        Builds a list of compilation units from a graph. Number of compilation units is maximized.
        """

        def __build_compilation_unit(
            graph: nx.DiGraph, start: Iterable[PurePath]
        ) -> CompilationUnit:
            nodes_subset = set()
            nodes_queue: deque[PurePath] = deque(start)
            versions: SolidityVersionRanges = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )

            while len(nodes_queue) > 0:
                node = nodes_queue.pop()
                versions &= graph.nodes[node]["versions"]

                if node in nodes_subset:
                    continue
                nodes_subset.add(node)

                for in_edge in graph.in_edges(node):
                    _from, to = in_edge
                    if _from not in nodes_subset:
                        nodes_queue.append(_from)

            subgraph = graph.subgraph(nodes_subset).copy()
            return CompilationUnit(subgraph, versions)

        sinks = [node for node, out_degree in graph.out_degree() if out_degree == 0]
        compilation_units = []

        for sink in sinks:
            compilation_unit = __build_compilation_unit(graph, [sink])
            compilation_units.append(compilation_unit)

        # cycles can also be "sinks" in terms of compilation units
        for cycle in nx.simple_cycles(graph):
            out_degree_sum = sum(
                out_degree for *_, out_degree in graph.out_degree(cycle)
            )

            if out_degree_sum == len(cycle):
                compilation_unit = __build_compilation_unit(graph, cycle)
                compilation_units.append(compilation_unit)
        return compilation_units

    @staticmethod
    def build_compilation_units_minimize(graph: nx.DiGraph) -> List[CompilationUnit]:
        """
        Builds a list of compilation units from a graph. Number of compilation units is *almost* minimized.
        This approach assures that every compiled file is in exactly one compilation unit.
        In very rare cases the project may not be possible to compile using this method. An example:
        - Lib.sol requires solc v0.5.*
        - A.sol requires solc v0.5.0 and imports Lib.sol
        - B.sol requires solc v0.5.1 and imports Lib.sol
        """
        compilation_units = []
        for component in nx.weakly_connected_components(graph):
            subgraph = graph.subgraph(component)

            versions: SolidityVersionRanges = SolidityVersionRanges(
                [SolidityVersionRange(None, None, None, None)]
            )
            for file in component:
                versions &= graph.nodes[file]["versions"]
            compilation_units.append(CompilationUnit(subgraph, versions))
        return compilation_units

    def create_build_settings(
        self, output_types: Collection[SolcOutputSelectionEnum]
    ) -> SolcInputSettings:
        settings = SolcInputSettings()  # type: ignore
        # TODO Allow setting all solc build settings
        # Currently it is not possible to set all solc standard JSON input build settings.
        # These include: stopAfter, optimizer, via_IR, debug, metadata, libraries and model checker settings.
        # See https://docs.soliditylang.org/en/v0.8.12/using-the-compiler.html#input-description.
        # Also it is not possible to specify solc output per contract or per source file.
        settings.remappings = [
            str(remapping) for remapping in self.__config.compiler.solc.remappings
        ]
        settings.evm_version = self.__config.compiler.solc.evm_version
        settings.via_IR = self.__config.compiler.solc.via_IR
        settings.optimizer = SolcInputOptimizerSettings(
            enabled=self.__config.compiler.solc.optimizer.enabled,
            runs=self.__config.compiler.solc.optimizer.runs,
        )

        settings.output_selection = {"*": {}}

        if SolcOutputSelectionEnum.ALL in output_types:
            settings.output_selection["*"][""] = [SolcOutputSelectionEnum.AST]  # type: ignore
            settings.output_selection["*"]["*"] = [SolcOutputSelectionEnum.ALL]  # type: ignore
        else:
            if SolcOutputSelectionEnum.AST in output_types:
                settings.output_selection["*"][""] = [SolcOutputSelectionEnum.AST]  # type: ignore
            settings.output_selection["*"]["*"] = [output_type for output_type in output_types if output_type != SolcOutputSelectionEnum.AST]  # type: ignore

        return settings

    def __write_global_artifacts(
        self,
        build_path: Path,
        build_settings: SolcInputSettings,
        output: Tuple[SolcOutput, ...],
        compilation_units: List[CompilationUnit],
        target_versions: List[SolidityVersion],
    ) -> None:
        units_info = {}

        # units are already sorted
        for index, (unit, target_version, out) in enumerate(
            zip(compilation_units, target_versions, output)
        ):
            sources = {}
            for source_unit_name in out.sources.keys():
                sources[source_unit_name] = (
                    Path(f"{index:03d}")
                    / "asts"
                    / sanitize_filename(
                        source_unit_name + ".json", "_", platform="universal"
                    )
                )

            contracts = {}
            for source_unit_name, info in out.contracts.items():
                contracts[source_unit_name] = {}
                for contract in info.keys():
                    contracts[source_unit_name][contract] = (
                        Path(f"{index:03d}")
                        / "contracts"
                        / sanitize_filename(
                            f"{source_unit_name}:{contract}.json",
                            "_",
                            platform="universal",
                        )
                    )

            info = CompilationUnitBuildInfo(
                build_dir=f"{index:03d}",
                sources=sources,
                contracts=contracts,
                errors=out.errors,
                source_units=unit.source_unit_names,
                allow_paths=self.__config.compiler.solc.allow_paths,
                include_paths=self.__config.compiler.solc.include_paths,
                settings=build_settings,
                compiler_version=target_version,
            )
            units_info[unit.hash.hex()] = info

        build_info = ProjectBuildInfo(compilation_units=units_info)
        with (build_path / "build.json").open("w") as f:
            f.write(build_info.json(by_alias=True, exclude_none=True))

    async def compile(
        self,
        files: Collection[Path],
        output_types: Collection[SolcOutputSelectionEnum],
        write_artifacts: bool = True,
        reuse_latest_artifacts: bool = True,
        modified_files: Optional[Mapping[Path, str]] = None,
        maximize_compilation_units: bool = False,
    ) -> List[Tuple[CompilationUnit, SolcOutput]]:
        if modified_files is None:
            modified_files = {}
        if len(files) + len(modified_files) == 0:
            raise CompilationError("No source files provided to compile.")
        target_version = self.__config.compiler.solc.target_version
        min_version = self.__config.min_solidity_version
        max_version = self.__config.max_solidity_version
        if target_version is not None and target_version < min_version:
            raise CompilationError(
                f"Target configured version {target_version} is lower than minimum supported version {min_version}"
            )
        if target_version is not None and target_version > max_version:
            raise CompilationError(
                f"Target configured version {target_version} is higher than maximum supported version {max_version}"
            )

        graph = self.build_graph(files, modified_files)
        if maximize_compilation_units:
            compilation_units = self.build_compilation_units_maximize(graph)
        else:
            compilation_units = self.build_compilation_units_minimize(graph)
        compilation_unit_hashes = {cu.hash for cu in compilation_units}
        if len(compilation_unit_hashes) != len(compilation_units):
            raise CompilationError(
                "Compilation units are not unique. This is a bug in the compiler."
            )
        build_settings = self.create_build_settings(output_types)

        # sort compilation units by their hash
        compilation_units.sort(key=lambda u: u.hash.hex())

        if write_artifacts:
            # prepare build dir
            build_path = self.__config.project_root_path / ".woke-build"
            if not build_path.is_dir():
                build_path.mkdir(parents=True, exist_ok=False)
            tmp_path = build_path / "tmp"
            if tmp_path.is_file():
                tmp_path.unlink()
            elif tmp_path.is_dir():
                shutil.rmtree(tmp_path)
        else:
            build_path = None

        latest_build_path = self.__config.project_root_path / ".woke-build"
        if reuse_latest_artifacts:
            try:
                latest_build_info = ProjectBuildInfo.parse_file(
                    latest_build_path / "build.json"
                )
            except (ValidationError, JSONDecodeError):
                logger.warning(
                    f"Failed to parse '{latest_build_path / 'build.json'}' file while trying to reuse the latest build artifacts."
                )
                latest_build_info = None
            except FileNotFoundError as e:
                logger.warning(
                    f"Unable to find '{e.filename}' file while trying to reuse the latest build artifacts."
                )
                latest_build_info = None
        else:
            latest_build_info = None

        target_versions = []
        for compilation_unit in compilation_units:
            target_version = self.__config.compiler.solc.target_version
            if target_version is not None:
                if target_version not in compilation_unit.versions:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"Unable to compile following files with solc version `{target_version}` set in config files:\n"
                        + files_str
                    )
            else:
                # use the latest matching version
                matching_versions = [
                    version
                    for version in reversed(self.__svm.list_all())
                    if version in compilation_unit.versions
                ]
                if len(matching_versions) == 0:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"Unable to compile following files with any solc version:\n"
                        + files_str
                    )
                try:
                    target_version = next(
                        version
                        for version in matching_versions
                        if version <= max_version
                    )
                except StopIteration:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"The maximum supported version of Solidity is {max_version}, unable to compile the following files:\n"
                        + files_str,
                    )
                if target_version < min_version:
                    files_str = "\n".join(str(path) for path in compilation_unit.files)
                    raise CompilationError(
                        f"The minimum supported version of Solidity is {min_version}, unable to compile the following files:\n"
                        + files_str,
                    )
            target_versions.append(target_version)

        for version in set(target_versions):
            if not self.__svm.installed(version):
                with Progress() as progress:
                    task = progress.add_task(
                        f"[green]Downloading solc {version}", total=1
                    )

                    async def on_progress(downloaded: int, total: int) -> None:
                        progress.update(task, completed=downloaded, total=total)  # type: ignore

                    await self.__svm.install(
                        version,
                        progress=on_progress,
                    )

        tasks = []
        for index, (compilation_unit, target_version) in enumerate(
            zip(compilation_units, target_versions)
        ):
            task = asyncio.create_task(
                self.__compile_unit(
                    compilation_unit,
                    target_version,
                    build_settings,
                    build_path / f"{index:03d}" if build_path is not None else None,
                    latest_build_info,
                )
            )
            tasks.append(task)

        # wait for compilation of all compilation units
        try:
            ret = await asyncio.gather(*tasks)
        except Exception:
            for task in tasks:
                task.cancel()
            raise
        if write_artifacts:
            if build_path is None:
                # should not really happen (it is present here just to silence the linter)
                raise ValueError("Build path is not set.")
            self.__rename_and_remove_artifacts(build_path)
            self.__write_global_artifacts(
                build_path, build_settings, ret, compilation_units, target_versions  # type: ignore
            )

        return [(cu, out) for cu, out in zip(compilation_units, ret)]

    @staticmethod
    def __rename_and_remove_artifacts(build_path: Path):
        """
        When this func is called .woke-build/ should the following contents:
          .woke-build/old-artifact-0,...,.woke-build/old-artifact-n      <- old artifacts
          .woke-build/tmp                                                <- tmp_dir with new artifacts
          .woke-build/build.json                                         <- old build info
        This func manages removal of the old artifacts and also moves the new ones out
        of the tmp/ directory.
        This is neccesary because old artifacts can't be removed before new ones are
        created and because we want to limit the amount of stored artifacts.
        """
        for fl in build_path.iterdir():
            if fl.stem != "tmp" and fl.stem != "build.json":
                if fl.is_dir():
                    shutil.rmtree(fl)
                else:
                    fl.unlink()
        if not (build_path / "tmp").exists():
            raise ValueError("New artificats do not exist.")
        for fl in (build_path / "tmp").iterdir():
            p = Path(fl).absolute()
            parent_dir = p.parents[1]
            p.rename(parent_dir / p.name)
        (build_path / "tmp").rmdir()

    async def __compile_unit(
        self,
        compilation_unit: CompilationUnit,
        target_version: SolidityVersion,
        build_settings: SolcInputSettings,
        build_path: Optional[Path],
        latest_build_info: Optional[ProjectBuildInfo],
    ) -> SolcOutput:
        # try to reuse the latest build artifacts
        if (
            latest_build_info is not None
            and compilation_unit.hash.hex() in latest_build_info.compilation_units
        ):
            latest_unit_info = latest_build_info.compilation_units[
                compilation_unit.hash.hex()
            ]

            if (
                latest_unit_info.source_units == compilation_unit.source_unit_names
                and latest_unit_info.allow_paths
                == self.__config.compiler.solc.allow_paths
                and latest_unit_info.include_paths
                == self.__config.compiler.solc.include_paths
                and latest_unit_info.settings == build_settings
                and latest_unit_info.compiler_version == target_version
            ):
                try:
                    logger.info("Reusing the latest build artifacts.")
                    latest_build_path = self.__config.project_root_path / ".woke-build"
                    sources = {}
                    for source, path in latest_unit_info.sources.items():
                        sources[source] = SolcOutputSourceInfo.parse_file(
                            latest_build_path / path
                        )

                    contracts = {}
                    for (
                        source_unit,
                        source_unit_info,
                    ) in latest_unit_info.contracts.items():
                        contracts[source_unit] = {}
                        for contract, path in source_unit_info.items():
                            contracts[source_unit][
                                contract
                            ] = SolcOutputContractInfo.parse_file(
                                latest_build_path / path
                            )
                    out = SolcOutput(
                        errors=latest_unit_info.errors,
                        sources=sources,
                        contracts=contracts,
                    )
                except (ValidationError, JSONDecodeError):
                    logger.warning(
                        "Failed to parse the latest build artifacts, falling back to solc compilation."
                    )
                    out = await self.compile_unit_raw(
                        compilation_unit, target_version, build_settings
                    )
                except FileNotFoundError as e:
                    logger.warning(
                        f"Unable to find '{e.filename}' file while reusing the latest build info. Build artifacts may be corrupted."
                    )
                    out = await self.compile_unit_raw(
                        compilation_unit, target_version, build_settings
                    )
            else:
                logger.info(
                    "Build settings have changed since the last build. Falling back to solc compilation."
                )
                out = await self.compile_unit_raw(
                    compilation_unit, target_version, build_settings
                )
        else:
            out = await self.compile_unit_raw(
                compilation_unit, target_version, build_settings
            )

        # write build artifacts
        if build_path is not None:
            tmp_path = build_path.parent / "tmp" / build_path.stem
            tmp_path.mkdir(parents=True, exist_ok=False)
            await self.__write_artifacts(out, tmp_path)

        return out

    async def compile_unit_raw(
        self,
        compilation_unit: CompilationUnit,
        target_version: SolidityVersion,
        build_settings: SolcInputSettings,
    ) -> SolcOutput:
        # Dict[source_unit_name: PurePath, path: Path]
        files = {}
        # Dict[source_unit_name: PurePath, content: str]
        sources = {}
        for source_unit_name, data in compilation_unit.graph.nodes.items():
            path = data["path"]
            content = data["content"]
            if content is None:
                files[source_unit_name] = path
            else:
                sources[source_unit_name] = content

        # run the solc executable
        return await self.__solc_frontend.compile(
            files, sources, target_version, build_settings
        )

    @staticmethod
    async def __write_artifacts(output: SolcOutput, build_path: Path) -> None:
        if output.sources is not None:
            ast_path = build_path / "asts"
            ast_path.mkdir(parents=False, exist_ok=False)
            for source_unit_name, value in output.sources.items():
                # AST output is generated per file (source unit name)
                # Because a source unit name can contain slashes, it is not possible to name the AST build file
                # by its source unit name. Blake2b hash of the Solidity source file content is used instead.
                file_path = ast_path / sanitize_filename(
                    source_unit_name + ".json", "_", platform="universal"
                )

                if file_path.is_file():
                    raise CompilationError(
                        f"Cannot write build info into '{file_path}' - file already exists."
                    )
                async with aiofiles.open(file_path, mode="w") as f:
                    await f.write(value.json(by_alias=True, exclude_none=True))

        if output.contracts is not None:
            contract_path = build_path / "contracts"
            contract_path.mkdir(parents=False, exist_ok=False)
            for source_unit_name, d in output.contracts.items():
                # All other build info is generated per contract. Contract names cannot contain slashes so it should
                # be safe to name the build info file by its contract name.
                for contract, info in d.items():
                    file_path = contract_path / sanitize_filename(
                        f"{source_unit_name}:{contract}.json", "_", platform="universal"
                    )

                    async with aiofiles.open(file_path, mode="w") as f:
                        await f.write(info.json(by_alias=True, exclude_none=True))



import pytest

from woke.core.solidity_version import (
    SolidityVersion,
    SolidityVersionExpr,
    SolidityVersionRange,
)


def test_version_basic_usage():
    v1 = SolidityVersion.fromstring("0.8.9-alpha.2+commit.12345678")
    assert v1.major == 0
    assert v1.minor == 8
    assert v1.patch == 9
    assert v1.prerelease == "alpha.2"
    assert v1.build == "commit.12345678"
    assert v1 < "0.8.13"

    v2 = SolidityVersion.fromstring("0.8.7")
    assert v1 > v2
    v3 = SolidityVersion.fromstring("0.8.9")
    assert v1 == v3
    v4 = SolidityVersion.fromstring("0.8.9-abc+def")
    assert v3 == v4  # prerelease and build tags are ignored


def test_version_str_and_repr():
    s = "1.2.3-abc.def-012-ABC-abc+xyz-123.XYZ"
    v = SolidityVersion.fromstring(s)
    assert str(v) == s
    assert eval(repr(v)) == v


def test_version_invalid():
    with pytest.raises(ValueError):
        SolidityVersion.fromstring(">0.8.1")
    with pytest.raises(ValueError):
        SolidityVersion.fromstring("=0.8.1")
    with pytest.raises(ValueError):
        SolidityVersion.fromstring("v0.8.1")
    with pytest.raises(ValueError):
        SolidityVersion.fromstring("x.8.1")


def test_version_range_basic():
    assert SolidityVersionRange(None, None, None, None) == SolidityVersionRange(
        "0.0.0", True, None, None
    )
    assert SolidityVersionRange("1.2.3", True, None, None) != SolidityVersionRange(
        "1.2.3", False, None, None
    )
    assert SolidityVersionRange(None, None, "3.4.5", True) != SolidityVersionRange(
        None, None, "3.4.5", False
    )

    r1 = SolidityVersionRange(None, None, None, None)
    assert not r1.isempty()
    assert r1.lower == SolidityVersion(0, 0, 0)
    assert r1.lower_inclusive
    assert r1.higher is None
    assert r1.higher_inclusive is None

    r2 = SolidityVersionRange("1.2.3", True, "3.4.5", False)
    assert not r2.isempty()
    assert r2.lower == SolidityVersion(1, 2, 3)
    assert r2.lower_inclusive
    assert r2.higher == SolidityVersion(3, 4, 5)
    assert not r2.higher_inclusive

    assert SolidityVersionRange("1.2.3", True, "0.9.9", False).isempty()
    assert SolidityVersionRange("1.2.3", True, "1.2.3", False).isempty()
    assert SolidityVersionRange("1.2.3", False, "1.2.3", True).isempty()
    assert SolidityVersionRange("1.2.3", False, "1.2.3", False).isempty()
    assert not SolidityVersionRange("1.2.3", True, "1.2.3", True).isempty()


def test_version_range_errors():
    r1 = SolidityVersionRange(None, None, None, None)
    with pytest.raises(ValueError):
        x = "abcd" in r1

    with pytest.raises(ValueError):
        SolidityVersionRange("-1.2.3", True, None, None)

    with pytest.raises(ValueError):
        SolidityVersionRange("1.2.3", None, None, None)
    with pytest.raises(ValueError):
        SolidityVersionRange(None, True, None, None)
    with pytest.raises(ValueError):
        SolidityVersionRange(None, None, "1.2.3", None)
    with pytest.raises(ValueError):
        SolidityVersionRange(None, None, None, True)


def test_version_range_contains():
    r1 = SolidityVersionRange("1.2.3", True, "2.0.0", False)
    assert SolidityVersion.fromstring("1.2.3") in r1
    assert "1.2.4" in r1
    assert "1.2.2" not in r1
    assert "2.0.0" not in r1
    assert "1.9.999" in r1

    r2 = SolidityVersionRange("0.8.9", False, "1.0.1", True)
    assert "0.8.9" not in r2
    assert "0.8.8" not in r2
    assert "0.8.10" in r2
    assert "1.0.1" in r2
    assert "1.0.0" in r2
    assert "0.9.9" in r2

    r3 = SolidityVersionRange("0.8.1", False, None, None)
    assert "0.8.1" not in r3
    assert "0.8.2" in r3
    assert "999999.999999.99999" in r3

    r4 = SolidityVersionRange("1.2.3", False, "0.1.2", False)
    assert r4.isempty()
    assert "0.0.0" not in r4
    assert "0.0.1" not in r4
    assert "0.1.2" not in r4
    assert "1.2.3" not in r4
    assert "1.2.4" not in r4


def test_version_range_str_and_repr():
    r1 = SolidityVersionRange(None, None, None, None)
    assert str(r1) == ">=0.0.0"
    assert eval(repr(r1)) == r1

    r2 = SolidityVersionRange("1.2.3", True, "4.5.6", False)
    assert str(r2) == ">=1.2.3 <4.5.6"
    assert eval(repr(r2)) == r2

    r3 = SolidityVersionRange("0.7.6", False, "2.0.7", True)
    assert str(r3) == ">0.7.6 <=2.0.7"
    assert eval(repr(r3)) == r3

    r4 = SolidityVersionRange("0.1.6", False, "0.0.8", True)
    assert r4.isempty()
    assert str(r4) == ">0.0.0 <0.0.0"
    assert eval(repr(r4)) == r4


def test_version_range_intersection():
    r1 = SolidityVersionRange("1.0.0", True, "2.0.0", True)
    r2 = SolidityVersionRange("1.0.1", False, None, None)
    assert r1 & r2 == SolidityVersionRange("1.0.1", False, "2.0.0", True)
    assert r1 & r2 == SolidityVersionRange.intersection(r1, r2)

    r3 = SolidityVersionRange("1.0.0", False, "2.0.0", False)
    assert r1 & r3 == r3
    assert r3 & r1 == r3
    assert r1 & r2 & r3 == r3 & r2 & r1
    assert r2 & r3 & r1 == SolidityVersionRange.intersection(r1, r3, r2)

    r4 = SolidityVersionRange("0.9.8", True, "1.9.8", False)
    assert r1 & r4 == SolidityVersionRange("1.0.0", True, "1.9.8", False)

    r5 = SolidityVersionRange("1.2.3", False, "2.0.1", False)
    assert r1 & r5 == SolidityVersionRange("1.2.3", False, "2.0.0", True)

    r6 = SolidityVersionRange(None, None, "1.0.0", False)
    r7 = SolidityVersionRange("2.0.0", False, None, None)
    assert (r1 & r6).isempty()
    assert r1 & r7

    r8 = SolidityVersionRange("0.0.0", False, "0.0.0", False)
    assert (r1 & r8).isempty()


def test_version_expr_invalid():
    invalid = [
        "v0.8.10",
        "v 0.8.10",
        ".1.2.3",
        "1.2.3.",
        "0.1.2.3.4",
        "abc",
        "o.8.7",
        "y.8.7",
        "1.2.3 - 4.5.6 - 7.8.9",
        ">=0.8.0 - 0.9.0",
        "0.8.0 - <0.9.0",
        "^1.2.3 - 4.5.6",
        "7.8.9 - ~1.2.4",
        "12.2.3 - x",
        "x.0.1",
        "x.0.x",
        "1.x.2",
        "x.x.2",
        "0.8.10-alpha.1",
        ">*",
        "<X",
        "<=x",
        "",
        "^x",
        "^0.0.0",
        "~*",
    ]
    for i in invalid:
        with pytest.raises(ValueError):
            SolidityVersionExpr(i)


def test_version_expr_comparators():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        ("=0.8.10", ["0.8.10"], ["0.8.9", "0.8.11", "0.0.0"]),
        ("0.2", ["0.2.0", "0.2.1", "0.2.7", "0.2.99"], ["0.1.999", "0.3.0", "0.0.0"]),
        ("=2.*.X", ["2.0.0", "2.0.1", "2.1.3"], ["1.2.3", "0.0.0", "3.1.0"]),
        ("*", ["0.0.0", "1.2.3", "0.2.8", "0.8.10"], []),
        (">=0.8.10", ["0.8.10", "0.8.11"], ["0.8.9"]),
        (">=1.2", ["1.2.0", "1.2.1", "1.3.0", "2.0.7"], ["1.1.9", "0.0.0", "1.1.1"]),
        (">=1.X.X", ["1.0.0", "1.2.3", "9.8.7"], ["0.9.9", "0.0.0"]),
        (">=*", ["0.0.0", "1.2.3", "0.3.4", "9.9.9"], []),
        (">0.8.10", ["0.8.11", "0.9.0", "1.0.1"], ["0.8.10", "0.8.9"]),
        (">0.6", ["0.7.0", "0.7.1", "1.2.3"], ["0.6.0", "0.6.999", "0.0.0"]),
        (">1", ["2.0.0", "2.0.1", "3.4.5"], ["1.0.0", "1.999.999"]),
        (
            "<0.8.10",
            ["0.8.9", "0.8.8", "0.7.0", "0.0.0"],
            ["0.8.10", "0.8.11", "0.9.0"],
        ),
        ("<1.1.x", ["1.0.9", "1.0.0", "0.8.9"], ["1.1.0", "1.1.1", "1.2.0"]),
        ("<1", ["0.9.9", "0.0.7", "0.9.99999"], ["1.0.0", "1.2.3", "2.0.1"]),
        (
            "<=0.8.10",
            ["0.8.10", "0.8.9", "0.5.1", "0.0.0"],
            ["0.8.11", "1.0.0", "0.9.0"],
        ),
        ("<=2.0", ["2.0.0", "2.0.999", "1.9.9"], ["2.1.0", "2.1.1", "3.0.1"]),
        ("<=1", ["1.0.0", "1.9.9", "0.8.9"], ["2.0.0", "2.8.1", "3.0.7"]),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"


def test_version_expr_hyphen():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        (
            "0.7.5 - 0.8.10",
            ["0.7.5", "0.8.1", "0.8.10"],
            ["0.7.4", "0.8.11", "0.0.0", "1.2.3"],
        ),
        ("1.2 - 1.3.2", ["1.2.0", "1.2.10", "1.3.0", "1.3.2"], ["1.1.999", "1.3.3"]),
        ("2 - 3", ["2.0.0", "2.5.7", "3.0.0", "3.9.9"], ["4.0.0", "4.9.99"]),
        (
            "0.2.5 - 0.3",
            ["0.2.5", "0.2.99", "0.3.1", "0.3.999"],
            ["0.2.4", "0.4.0", "0.4.2"],
        ),
        (
            "0.6.9 - 1",
            ["0.6.9", "0.9.9", "1.0.0", "1.8.7"],
            ["0.6.8", "2.0.0", "0.0.0"],
        ),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"


def test_version_expr_tilde():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        ("~0.8.7", ["0.8.7", "0.8.8", "0.8.999"], ["0.9.0", "0.8.6"]),
        ("~2.1", ["2.1.0", "2.1.1", "2.1.999"], ["2.2.0", "2.2.1", "2.0.9"]),
        ("~1", ["1.0.0", "1.2.3", "1.999.99"], ["2.0.0", "2.0.2", "0.9.9"]),
        ("~0", ["0.0.0", "0.1.2", "0.9.9"], ["1.0.0", "2.1.3"]),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"


def test_version_expr_caret():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        ("^3.2.1", ["3.2.1", "3.9.8", "3.5.0"], ["3.2.0", "4.0.0"]),
        ("^0.4.0", ["0.4.0", "0.4.1", "0.4.99"], ["0.3.99", "0.5.0"]),
        ("^0.0.7", ["0.0.7"], ["0.0.6", "0.0.8", "0.0.0"]),
        ("^1.2", ["1.2.0", "1.9.99"], ["2.0.0", "1.1.99"]),
        ("^0.0.X", ["0.0.0", "0.0.99"], ["0.1.0", "1.0.0"]),
        ("^1", ["1.0.0", "1.9.8", "1.4.0"], ["0.0.0", "0.9.9", "2.0.0"]),
        ("^0.*.X", ["0.0.0", "0.9.70"], ["1.0.0", "1.2.0"]),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"


def test_version_expr_complex():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        (
            "0.8.7 || 1 - 1.2.7",
            ["0.8.7", "1.0.0", "1.2.7"],
            ["0.8.6", "0.8.8", "0.9.9", "1.2.8"],
        ),
        (
            "^0.8 || 0.6.1 - 0.7.8",
            ["0.6.1", "0.7.0", "0.7.8", "0.8.0", "0.8.9"],
            ["0.6.0", "0.7.9", "0.9.0"],
        ),
        ("~0 || >=1.0.0 <1 || ^0", ["0.0.0", "0.1.2", "0.9.1"], ["1.0.0", "1.2.3"]),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"


def test_version_expr_whitespace():
    expressions = [
        # expression, list of matching versions, list of nonmatching versions
        (
            "      0 .\t8\n.\r7||1 - 1.2.7\n",
            ["0.8.7", "1.0.0", "1.2.7"],
            ["0.8.6", "0.8.8", "0.9.9", "1.2.8"],
        ),
        (
            "\r^0.8\t||0.6.1 - 0.7.8\r",
            ["0.6.1", "0.7.0", "0.7.8", "0.8.0", "0.8.9"],
            ["0.6.0", "0.7.9", "0.9.0"],
        ),
        (
            "~\t0\n \t|| >=\r\n1.0.0<1||^0",
            ["0.0.0", "0.1.2", "0.9.1"],
            ["1.0.0", "1.2.3"],
        ),
    ]

    for exp in expressions:
        e = SolidityVersionExpr(exp[0])
        for matching in exp[1]:
            version = SolidityVersion.fromstring(matching)
            assert (
                version in e
                and version in e.version_ranges
                and matching in e
                and matching in e.version_ranges
            ), f"Assertion failed: {matching} in {exp[0]}"
        for nonmatching in exp[2]:
            version = SolidityVersion.fromstring(nonmatching)
            assert (
                version not in e
                and version not in e.version_ranges
                and nonmatching not in e
                and nonmatching not in e.version_ranges
            ), f"Assertion failed: {nonmatching} not in {exp[0]}"



import asyncio
import os
import platform
import shutil
import stat
import subprocess
from pathlib import Path

import pytest
from click.testing import CliRunner
from git import Repo  # type: ignore

from woke.cli.__main__ import main
from woke.compile import SolcOutputSelectionEnum, SolidityCompiler
from woke.config import WokeConfig
from woke.utils import change_cwd

PYTEST_BUILD_PATH = Path.home() / ".tmpwoke_rkDv61DDf7"


@pytest.fixture()
def config():
    config_dict = {"compiler": {"solc": {"include_paths": ["./node_modules"]}}}
    return WokeConfig.fromdict(
        config_dict,
        woke_root_path=PYTEST_BUILD_PATH,
        project_root_path=PYTEST_BUILD_PATH,
    )


@pytest.fixture()
def setup_project(request):
    clone_url = request.param
    repo = None

    try:
        repo = Repo.clone_from(
            clone_url, PYTEST_BUILD_PATH, multi_options=["--depth=1"]
        )
        (PYTEST_BUILD_PATH / "woke.toml").write_text(
            '[compiler.solc]\ninclude_paths = ["./node_modules"]'
        )
        (PYTEST_BUILD_PATH / "woke_root").mkdir()
        subprocess.run(
            ["yarn", "install"],
            cwd=PYTEST_BUILD_PATH,
            shell=(platform.system() == "Windows"),
        )

        yield
    finally:

        def onerror(func, path, exc_info):
            if not os.access(path, os.W_OK):
                os.chmod(path, stat.S_IWUSR)
                func(path)
            else:
                raise

        if repo is not None:
            repo.close()
        shutil.rmtree(PYTEST_BUILD_PATH, onerror=onerror)


@pytest.mark.slow
@pytest.mark.parametrize(
    "setup_project", [r"https://github.com/Uniswap/v3-core.git"], indirect=True
)
def test_compile_uniswap_v3(setup_project, config):
    files = list((PYTEST_BUILD_PATH / "contracts").rglob("*.sol"))
    compiler = SolidityCompiler(config)
    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=False
        )
    )
    assert len(output)

    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=True
        )
    )
    assert len(output)

    cli_runner = CliRunner()
    with change_cwd(PYTEST_BUILD_PATH):
        cli_result = cli_runner.invoke(
            main, [f"--woke-root-path={PYTEST_BUILD_PATH / 'woke_root'}", "compile"]
        )
    assert cli_result.exit_code == 0


@pytest.mark.slow
@pytest.mark.parametrize(
    "setup_project", [r"https://github.com/graphprotocol/contracts.git"], indirect=True
)
def test_compile_the_graph(setup_project, config):
    files = list((PYTEST_BUILD_PATH / "contracts").rglob("*.sol"))
    compiler = SolidityCompiler(config)
    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=False
        )
    )
    assert len(output)

    compiler = SolidityCompiler(config)
    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=True
        )
    )
    assert len(output)

    cli_runner = CliRunner()
    with change_cwd(PYTEST_BUILD_PATH):
        cli_result = cli_runner.invoke(
            main, [f"--woke-root-path={PYTEST_BUILD_PATH / 'woke_root'}", "compile"]
        )
    assert cli_result.exit_code == 0


@pytest.mark.slow
@pytest.mark.parametrize(
    "setup_project", [r"https://github.com/traderjoe-xyz/joe-core.git"], indirect=True
)
def test_compile_trader_joe(setup_project, config):
    files = list((PYTEST_BUILD_PATH / "contracts").rglob("*.sol"))
    compiler = SolidityCompiler(config)
    output = asyncio.run(
        compiler.compile(
            files,
            [SolcOutputSelectionEnum.ALL],
            reuse_latest_artifacts=False,
            maximize_compilation_units=True,
        )
    )
    assert len(output)

    output = asyncio.run(
        compiler.compile(
            files,
            [SolcOutputSelectionEnum.ALL],
            reuse_latest_artifacts=True,
            maximize_compilation_units=True,
        )
    )
    assert len(output)

    cli_runner = CliRunner()
    with change_cwd(PYTEST_BUILD_PATH):
        cli_result = cli_runner.invoke(
            main, [f"--woke-root-path={PYTEST_BUILD_PATH / 'woke_root'}", "compile"]
        )
    assert cli_result.exit_code == 0


@pytest.mark.slow
@pytest.mark.parametrize(
    "setup_project",
    [r"https://github.com/axelarnetwork/axelar-cgp-solidity.git"],
    indirect=True,
)
def test_compile_axelar(setup_project, config):
    files = list((PYTEST_BUILD_PATH / "contracts").rglob("*.sol"))
    compiler = SolidityCompiler(config)
    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=False
        )
    )
    assert len(output)

    output = asyncio.run(
        compiler.compile(
            files, [SolcOutputSelectionEnum.ALL], reuse_latest_artifacts=True
        )
    )
    assert len(output)

    cli_runner = CliRunner()
    with change_cwd(PYTEST_BUILD_PATH):
        cli_result = cli_runner.invoke(
            main,
            [f"--woke-root-path={PYTEST_BUILD_PATH / 'woke_root'}", "compile"]
            + [str(file.resolve()) for file in files],
        )
    assert cli_result.exit_code == 0



import inspect
from typing import Callable, Optional, Tuple, TypeVar

from .campaign import logger

ModelUpdater = Callable[..., Optional[Tuple]]
TxnReturn = TypeVar("TxnReturn")
Txn = Callable[..., TxnReturn]
TxnEnhanced = Callable[..., TxnReturn]
Decorator = Callable[..., Txn]


def model(model_updater: ModelUpdater) -> Decorator:
    def decorator(txn_enhanced: TxnEnhanced) -> Txn:
        def txn(*args, **kwars) -> TxnReturn:  # type: ignore
            m_res = model_updater(*args)
            if m_res is None:
                return txn_enhanced(*args)  # type: ignore
            else:
                t_names = inspect.signature(txn_enhanced).parameters.keys()
                m_names = inspect.signature(model_updater).parameters.keys()
                # Sanity check
                assert len(m_res) == len(t_names) - len(m_names)
                for m, t in zip(m_names, t_names):
                    assert m == t
                for idx, p in enumerate(tuple(t_names)[len(m_names) :]):
                    logger.debug(f"{p} = {m_res[idx]}")
                return txn_enhanced(*args, *m_res)  # type: ignore

        return txn  # type:ignore

    return decorator


def ignore(fn):
    fn.ignore = True
    return fn


def flow(fn):
    fn.flow = True
    return fn


def invariant(fn):
    fn.invariant = True
    return fn


def precondition(_precondition):
    def decorator(fn):
        fn.precondition = _precondition
        return fn

    return decorator


def max_times(x: int):
    def decorator(fn):
        fn.max_times = x
        return fn

    return decorator


def weight(w: int):
    def decorator(fn):
        fn.weight = w
        return fn

    return decorator



import re
import shutil
from pathlib import Path

import pytest
from click.testing import CliRunner

from woke.cli.__main__ import main
from woke.config import WokeConfig
from woke.core.solidity_version import SolidityVersion
from woke.svm import SolcVersionManager

PYTEST_WOKE_ROOT_PATH = (Path.home() / ".tmpwoke_Z6yVySfqSk").resolve()


@pytest.fixture()
def woke_root_path():
    PYTEST_WOKE_ROOT_PATH.mkdir()
    yield
    shutil.rmtree(PYTEST_WOKE_ROOT_PATH, True)


@pytest.mark.slow
def test_svm_install(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    svm = SolcVersionManager(config)
    cli_runner = CliRunner()

    latest_version = next(reversed(svm.list_all()))
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "*"]
    )
    print(cli_result.output)
    assert cli_result.exit_code == 0
    assert latest_version in svm.list_installed()


@pytest.mark.slow
def test_svm_switch(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    svm = SolcVersionManager(config)
    cli_runner = CliRunner()

    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "switch", "0.6.2"]
    )
    assert cli_result.exit_code != 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.6.2"]
    )
    assert cli_result.exit_code == 0
    assert len(svm.list_installed()) == 1
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "switch", "0.6.2"]
    )
    assert cli_result.exit_code == 0

    version_str = (PYTEST_WOKE_ROOT_PATH / ".woke_solc_version").read_text()
    version = SolidityVersion.fromstring(version_str)
    assert version in svm.list_installed()
    assert version == "0.6.2"


@pytest.mark.slow
def test_svm_use(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    svm = SolcVersionManager(config)
    cli_runner = CliRunner()

    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "use", "0.8.4"]
    )
    assert cli_result.exit_code == 0
    assert len(svm.list_installed()) == 1

    version_str = (PYTEST_WOKE_ROOT_PATH / ".woke_solc_version").read_text()
    version = SolidityVersion.fromstring(version_str)
    assert version in svm.list_installed()
    assert version == "0.8.4"


@pytest.mark.slow
def test_svm_list(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    cli_runner = CliRunner()

    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.7.6"]
    )
    assert cli_result.exit_code == 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.8.0"]
    )
    assert cli_result.exit_code == 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "list"]
    )
    assert cli_result.exit_code == 0

    assert re.search(r"^.*\s0\.7\.6\s*$", cli_result.output, re.MULTILINE)
    assert re.search(r"^.*\s0\.8\.0\s*$", cli_result.output, re.MULTILINE)


@pytest.mark.slow
def test_svm_list_all(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    cli_runner = CliRunner()

    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.4.20"]
    )
    assert cli_result.exit_code == 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.6.3"]
    )
    assert cli_result.exit_code == 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "list", "--all"]
    )
    assert cli_result.exit_code == 0

    assert re.search(r"^.*\s0\.4\.19\s*$", cli_result.output, re.MULTILINE)
    assert re.search(r"^.*\s0\.4\.20\s.*installed.*$", cli_result.output, re.MULTILINE)
    assert re.search(r"^.*\s0\.5\.4\s*$", cli_result.output, re.MULTILINE)
    assert re.search(r"^.*\s0\.6\.3\s.*installed.*$", cli_result.output, re.MULTILINE)
    assert re.search(r"^.*\s0\.8\.12\s*$", cli_result.output, re.MULTILINE)


@pytest.mark.slow
def test_svm_remove(woke_root_path):
    config = WokeConfig(woke_root_path=PYTEST_WOKE_ROOT_PATH)
    config.load_configs()
    svm = SolcVersionManager(config)
    cli_runner = CliRunner()

    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "remove", "0.6.5"]
    )
    assert cli_result.exit_code != 0
    cli_result = cli_runner.invoke(
        main,
        [
            f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}",
            "svm",
            "remove",
            "0.6.5",
            "--ignore-missing",
        ],
    )
    assert cli_result.exit_code == 0
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "install", "0.6.3"]
    )
    assert cli_result.exit_code == 0
    assert "0.6.3" in svm.list_installed()
    cli_result = cli_runner.invoke(
        main, [f"--woke-root-path={PYTEST_WOKE_ROOT_PATH}", "svm", "remove", "0.6.3"]
    )
    assert cli_result.exit_code == 0
    assert len(svm.list_installed()) == 0
    assert not svm.get_path("0.6.3").parent.is_dir()



import pytest

from woke.regex_parsing.solidity_import import SolidityImportExpr


def test_import_simple():
    assert SolidityImportExpr("'filename'").filename == "filename"
    assert SolidityImportExpr("*as symbolName from'filename'").filename == "filename"
    assert SolidityImportExpr("'filename'as symbolName").filename == "filename"
    assert (
        SolidityImportExpr("{symbol1 as alias,symbol2}from'filename'").filename
        == "filename"
    )


def test_import_whitespace():
    assert SolidityImportExpr("\r' \t filename'").filename == " \t filename"
    assert (
        SolidityImportExpr("\n*\ras\tsymbolName\r\n from '  f\tilename'").filename
        == "  f\tilename"
    )
    assert SolidityImportExpr("'filename\t'\tas\nsymbolName").filename == "filename\t"
    assert (
        SolidityImportExpr(
            "{\r\nsymbol1\ras   alias  \r,\t  \r\nsymbol2\n}\r\n \tfrom\n\r' filename '"
        ).filename
        == " filename "
    )


def test_import_escape():
    filename1 = r"""'\'filename\"'"""
    filename2 = r'''"\"filename\'"'''
    assert (
        SolidityImportExpr("{filename}".format(filename=filename1)).filename
        == "'filename\""
    )
    assert (
        SolidityImportExpr(
            "*as symbolName from {filename}".format(filename=filename2)
        ).filename
        == "\"filename'"
    )
    assert (
        SolidityImportExpr(
            "{filename}as symbolName".format(filename=filename1)
        ).filename
        == "'filename\""
    )
    assert (
        SolidityImportExpr(
            "{{symbol1 as alias,symbol2}}from{filename}".format(filename=filename2)
        ).filename
        == "\"filename'"
    )


def test_import_invalid():
    with pytest.raises(ValueError):
        SolidityImportExpr("'file\nname'")
    with pytest.raises(ValueError):
        SolidityImportExpr("*as symbolName from'f\nilename'")
    with pytest.raises(ValueError):
        SolidityImportExpr("* from 'abc.sol'")



import inspect
import logging
import multiprocessing
import multiprocessing.connection
import multiprocessing.synchronize
import os
import pickle
import random
import sys
import types
from contextlib import closing, redirect_stderr, redirect_stdout
from pathlib import Path
from typing import Callable, Iterable

import brownie
import rich.progress
from brownie import rpc, web3
from brownie._config import CONFIG
from brownie.test.managers.runner import RevertContextManager
from ipdb.__main__ import _init_pdb
from IPython.core.debugger import BdbQuit_excepthook
from IPython.utils.io import Tee
from pathvalidate import sanitize_filename  # type: ignore
from rich.traceback import Traceback
from tblib import pickling_support

from woke.cli.console import console
from woke.config import WokeConfig


def _setup(port: int, network_id: str) -> None:
    brownie.reverts = RevertContextManager
    active_network = CONFIG.set_active_network(network_id)

    # Disables color formatting for brownie output
    CONFIG.settings["console"]["show_colors"] = False

    web3.connect(f"{active_network['host']}:{port}")

    cmd_settings = active_network["cmd_settings"]
    cmd_settings["port"] = port

    rpc.launch(active_network["cmd"], **cmd_settings)


def _attach_debugger() -> None:
    # TODO Implement `ipdb` package functionalities
    # This function relies on `ipdb` internal functions
    # We could implement all the `ipdb` functionalities ourselves (it is rather small codebase)
    if sys.excepthook != BdbQuit_excepthook:
        BdbQuit_excepthook.excepthook_ori = sys.excepthook
        sys.excepthook = BdbQuit_excepthook
    p = _init_pdb(
        commands=["import IPython", "alias embed() IPython.embed(colors='neutral')"]
    )
    p.reset()
    p.interaction(None, sys.exc_info()[2])


def _run_core(
    fuzz_test: Callable,
    index: int,
    port: int,
    random_seed: bytes,
    finished_event: multiprocessing.synchronize.Event,
    child_conn: multiprocessing.connection.Connection,
    network_id: str,
):
    print(f"Using seed '{random_seed.hex()}' for process #{index}")
    _setup(port, network_id)

    project = brownie.project.load()

    brownie.chain.reset()

    args = []
    for arg in inspect.getfullargspec(fuzz_test).args:
        if arg in {"a", "accounts"}:
            args.append(brownie.accounts)
        elif arg == "chain":
            args.append(brownie.chain)
        elif arg == "Contract":
            args.append(brownie.Contract)
        elif arg == "history":
            args.append(brownie.history)
        elif arg == "interface":
            args.append(project.interface)
        elif arg == "rpc":
            args.append(brownie.rpc)
        elif arg == "web3":
            args.append(brownie.web3)
        elif arg in project.keys():
            args.append(project[arg])
        else:
            raise ValueError(
                f"Unable to set value for '{arg}' argument in '{fuzz_test.__name__}' function."
            )
    fuzz_test(*args)

    child_conn.send(None)
    finished_event.set()


def _run(
    fuzz_test: Callable,
    index: int,
    port: int,
    random_seed: bytes,
    log_file: Path,
    tee: bool,
    finished_event: multiprocessing.synchronize.Event,
    child_conn: multiprocessing.connection.Connection,
    network_id: str,
):
    pickling_support.install()
    random.seed(random_seed)

    if not tee:
        logging.basicConfig(filename=log_file)

    try:
        if tee:
            with closing(Tee(log_file)):
                _run_core(
                    fuzz_test,
                    index,
                    port,
                    random_seed,
                    finished_event,
                    child_conn,
                    network_id,
                )
        else:
            with log_file.open("w") as f, redirect_stdout(f), redirect_stderr(f):
                _run_core(
                    fuzz_test,
                    index,
                    port,
                    random_seed,
                    finished_event,
                    child_conn,
                    network_id,
                )
    except Exception:
        child_conn.send(pickle.dumps(sys.exc_info()))
        finished_event.set()

        try:
            attach: bool = child_conn.recv()
            if attach:
                sys.stdin = os.fdopen(0)
                _attach_debugger()
        finally:
            finished_event.set()
    finally:
        with log_file.open("a") as f, redirect_stdout(f), redirect_stderr(f):
            rpc.kill()


def fuzz(
    config: WokeConfig,
    fuzz_test: types.FunctionType,
    process_count: int,
    seeds: Iterable[bytes],
    logs_dir: Path,
    passive: bool,
    network_id: str,
):
    random_seeds = list(seeds)
    if len(random_seeds) < process_count:
        for i in range(process_count - len(random_seeds)):
            random_seeds.append(os.urandom(8))

    processes = dict()
    for i, seed in zip(range(process_count), random_seeds):
        console.print(f"Using seed '{seed.hex()}' for process #{i}")
        finished_event = multiprocessing.Event()
        parent_conn, child_conn = multiprocessing.Pipe()

        log_path = logs_dir / sanitize_filename(
            f"{fuzz_test.__module__}.{fuzz_test.__name__}_{i}.ansi"
        )

        p = multiprocessing.Process(
            target=_run,
            args=(
                fuzz_test,
                i,
                8545 + i,
                seed,
                log_path,
                passive and i == 0,
                finished_event,
                child_conn,
                network_id,
            ),
        )
        processes[i] = (p, finished_event, parent_conn, child_conn)
        p.start()

    with rich.progress.Progress(
        rich.progress.SpinnerColumn(finished_text="[green]⠿"),
        "[progress.description][yellow]{task.description}, "
        "[green]{task.fields[thr_rem]}[yellow] "
        "processes remaining",
    ) as progress:
        if passive:
            progress.stop()
        task = progress.add_task("Fuzzing", thr_rem=len(processes), total=1)

        while len(processes):
            to_be_removed = []
            for i, (p, e, parent_conn, child_conn) in processes.items():
                finished = e.wait(0.125)
                if finished:
                    to_be_removed.append(i)

                    exception_info = parent_conn.recv()
                    if exception_info is not None:
                        exception_info = pickle.loads(exception_info)

                    if exception_info is not None:
                        if not passive or i == 0:
                            tb = Traceback.from_exception(
                                exception_info[0], exception_info[1], exception_info[2]
                            )

                            if not passive:
                                progress.stop()

                            console.print(tb)
                            console.print(
                                f"Process #{i} failed with an exception above."
                            )

                            attach = None
                            while attach is None:
                                response = input(
                                    "Would you like to attach the debugger? [y/n] "
                                )
                                if response == "y":
                                    attach = True
                                elif response == "n":
                                    attach = False
                        else:
                            attach = False

                        e.clear()
                        parent_conn.send(attach)
                        e.wait()
                        if not passive:
                            progress.start()

                    progress.update(
                        task, thr_rem=len(processes) - len(to_be_removed)
                    )
                    if i == 0:
                        progress.start()
            for i in to_be_removed:
                processes.pop(i)
        progress.update(task, description="Finished", completed=1)



from pathlib import Path, PurePath, PurePosixPath, PureWindowsPath
from typing import Dict, FrozenSet, List

from pydantic import BaseModel, Extra, validator

from woke.compile.solc_frontend import SolcInputSettings, SolcOutputError
from woke.core.solidity_version import SolidityVersion


class BuildInfoModel(BaseModel):
    class Config:
        extra = Extra.allow
        allow_mutation = False
        arbitrary_types_allowed = True
        json_encoders = {PurePosixPath: str, PureWindowsPath: str, SolidityVersion: str}


class CompilationUnitBuildInfo(BuildInfoModel):
    build_dir: str  # TODO unused
    sources: Dict[str, Path]
    contracts: Dict[str, Dict[str, Path]]
    errors: List[SolcOutputError]
    source_units: FrozenSet[PurePath]
    allow_paths: FrozenSet[Path]
    include_paths: FrozenSet[Path]
    settings: SolcInputSettings
    compiler_version: SolidityVersion

    @validator("source_units", pre=True, each_item=True)
    def set_source_units(cls, v):
        return PurePath(v)


class ProjectBuildInfo(BuildInfoModel):
    compilation_units: Dict[str, CompilationUnitBuildInfo]



from __future__ import annotations

import heapq
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Callable, DefaultDict, Dict, List, Tuple, Union

from intervaltree import IntervalTree

from woke.ast.enums import GlobalSymbolsEnum

if TYPE_CHECKING:
    from woke.ast.ir.abc import SolidityAbc
    from woke.ast.ir.expression.identifier import Identifier
    from woke.ast.ir.expression.member_access import MemberAccess
    from woke.ast.ir.meta.source_unit import SourceUnit

from woke.ast.nodes import AstNodeId, AstSolc

logger = logging.getLogger(__name__)


@dataclass
class CallbackParams:
    interval_trees: Dict[Path, IntervalTree]
    source_units: Dict[Path, SourceUnit]


@dataclass(order=True)
class PostProcessQueueItem:
    priority: int
    callback: Callable[[CallbackParams], None] = field(compare=False)


class ReferenceResolver:
    __ordered_nodes: Dict[bytes, Dict[AstNodeId, Tuple[Path, int]]]
    __registered_nodes: Dict[Tuple[Path, int], SolidityAbc]
    __post_process_callbacks: List[PostProcessQueueItem]
    __destroy_callbacks: DefaultDict[Path, List[Callable[[], None]]]
    __global_symbol_references: DefaultDict[
        GlobalSymbolsEnum, List[Union[Identifier, MemberAccess]]
    ]

    def __init__(self):
        self.__ordered_nodes = defaultdict(dict)
        self.__registered_nodes = {}
        self.__post_process_callbacks = []
        self.__destroy_callbacks = defaultdict(list)
        self.__global_symbol_references = defaultdict(list)

    def index_nodes(self, root_node: AstSolc, path: Path, cu_hash: bytes) -> None:
        self.__ordered_nodes[cu_hash][root_node.id] = (path, 0)
        for index, node in enumerate(root_node):
            self.__ordered_nodes[cu_hash][node.id] = (path, index + 1)

    def get_node_path_order(
        self, node_id: AstNodeId, cu_hash: bytes
    ) -> Tuple[Path, int]:
        return self.__ordered_nodes[cu_hash][node_id]

    def get_ast_id_from_cu_node_path_order(
        self, node_path_order: Tuple[Path, int], cu_hash: bytes
    ) -> AstNodeId:
        for node_id, node_path_order_ in self.__ordered_nodes[cu_hash].items():
            if node_path_order_ == node_path_order:
                return node_id
        raise KeyError(
            f"No node found for path order {node_path_order} cu hash {cu_hash.hex()}"
        )

    def register_node(self, node: SolidityAbc, node_id: AstNodeId, cu_hash: bytes):
        assert cu_hash in self.__ordered_nodes
        assert node_id in self.__ordered_nodes[cu_hash]
        node_path_order = self.__ordered_nodes[cu_hash][node_id]
        self.__registered_nodes[node_path_order] = node

    def resolve_node(self, node_id: AstNodeId, cu_hash: bytes) -> SolidityAbc:
        node_path_order = self.__ordered_nodes[cu_hash][node_id]
        return self.__registered_nodes[node_path_order]

    def register_post_process_callback(
        self, callback: Callable[[CallbackParams], None], priority: int = 0
    ):
        heapq.heappush(
            self.__post_process_callbacks, PostProcessQueueItem(priority, callback)
        )

    def register_destroy_callback(self, file: Path, callback: Callable[[], None]):
        self.__destroy_callbacks[file].append(callback)

    def run_post_process_callbacks(self, callback_params: CallbackParams):
        while len(self.__post_process_callbacks):
            callback = heapq.heappop(self.__post_process_callbacks).callback
            callback(callback_params)

    def run_destroy_callbacks(self, file: Path):
        for callback in self.__destroy_callbacks[file]:
            callback()
        del self.__destroy_callbacks[file]

    def register_global_symbol_reference(
        self, node_id: GlobalSymbolsEnum, node: Union[Identifier, MemberAccess]
    ):
        self.__global_symbol_references[node_id].append(node)

    def unregister_global_symbol_reference(
        self, node_id: GlobalSymbolsEnum, node: Union[Identifier, MemberAccess]
    ):
        self.__global_symbol_references[node_id].remove(node)

    def get_global_symbol_references(
        self, node_id: GlobalSymbolsEnum
    ) -> Tuple[Union[Identifier, MemberAccess], ...]:
        return tuple(self.__global_symbol_references[node_id])
